from __future__ import annotations

import logging
import typing as t

from globus_sdk._internal import guards

log = logging.getLogger(__name__)


class ErrorInfo:
    """
    Errors may contain "containers" of data which are testable (define ``__bool__``).
    When they have data, they should ``bool()`` as ``True``
    """

    _has_data: bool

    def __bool__(self) -> bool:
        return self._has_data

    def __str__(self) -> str:
        if self:
            attrmap = ", ".join(
                [f"{k}={v}" for k, v in self.__dict__.items() if not k.startswith("_")]
            )
        else:
            attrmap = ":"
        return f"{self.__class__.__name__}({attrmap})"


class AuthorizationParameterInfo(ErrorInfo):
    """
    AuthorizationParameterInfo objects may contain information about the
    'authorization_parameters' of an error. They test as truthy when the error has valid
    'authorization_parameters' data.

    :ivar session_message: A message from the server
    :vartype session_message: str, optional

    :ivar session_required_identities: A list of identity IDs as strings which are being
        requested by the server
    :vartype session_required_identities: list of str, optional

    :ivar session_required_single_domain: A list of domains which are being requested by
        the server ("single domain" because the user should choose one)
    :vartype session_required_single_domain: list of str, optional

    :ivar session_required_policies: A list of policies required for the session.
    :vartype session_required_policies: list of str, optional

    :ivar session_required_mfa: Whether MFA is required for the session.
    :vartype session_required_mfa: bool, optional

    **Examples**

    >>> try:
    >>>     ...  # something
    >>> except GlobusAPIError as err:
    >>>     # get a parsed AuthorizationParameterInfo object, and check if it's truthy
    >>>     authz_params = err.info.authorization_parameters
    >>>     if not authz_params:
    >>>         raise
    >>>     # whatever handling code is desired...
    >>>     print("got authz params:", authz_params)
    """

    def __init__(self, error_data: dict[str, t.Any]) -> None:
        # data is there if this key is present and it is a dict
        self._has_data = isinstance(error_data.get("authorization_parameters"), dict)
        data = t.cast(
            t.Dict[str, t.Any], error_data.get("authorization_parameters", {})
        )

        self.session_message: str | None = self._parse_session_message(data)

        self.session_required_identities: list[str] | None = (
            self._parse_session_required_identities(data)
        )

        self.session_required_single_domain: list[str] | None = (
            self._parse_session_required_single_domain(data)
        )

        self.session_required_policies: list[str] | None = (
            self._parse_session_required_policies(data)
        )

        self.session_required_mfa = self._parse_session_required_mfa(data)

    def _parse_session_message(self, data: dict[str, t.Any]) -> str | None:
        session_message = data.get("session_message")
        if isinstance(session_message, str):
            return session_message
        elif session_message is not None:
            self._warn_type("session_message", "str", session_message)
        return None

    def _parse_session_required_identities(
        self, data: dict[str, t.Any]
    ) -> list[str] | None:
        session_required_identities = data.get("session_required_identities")
        if guards.is_list_of(session_required_identities, str):
            return session_required_identities
        elif session_required_identities is not None:
            self._warn_type(
                "session_required_identities",
                "list[str]",
                session_required_identities,
            )
        return None

    def _parse_session_required_single_domain(
        self, data: dict[str, t.Any]
    ) -> list[str] | None:
        session_required_single_domain = data.get("session_required_single_domain")
        if guards.is_list_of(session_required_single_domain, str):
            return session_required_single_domain
        elif session_required_single_domain is not None:
            self._warn_type(
                "session_required_single_domain",
                "list[str]",
                session_required_single_domain,
            )
        return None

    def _parse_session_required_policies(
        self, data: dict[str, t.Any]
    ) -> list[str] | None:
        session_required_policies = data.get("session_required_policies")
        if isinstance(session_required_policies, str):
            return session_required_policies.split(",")
        elif guards.is_list_of(session_required_policies, str):
            return session_required_policies
        elif session_required_policies is not None:
            self._warn_type(
                "session_required_policies", "list[str]|str", session_required_policies
            )
        return None

    def _parse_session_required_mfa(self, data: dict[str, t.Any]) -> bool | None:
        session_required_mfa = data.get("session_required_mfa")
        if isinstance(session_required_mfa, bool):
            return session_required_mfa
        elif session_required_mfa is not None:
            self._warn_type("session_required_mfa", "bool", session_required_mfa)
        return None

    def _warn_type(self, key: str, expected: str, got: t.Any) -> None:
        log.warning(
            f"During ErrorInfo instantiation, got unexpected type for '{key}'. "
            f"Expected '{expected}'. Got '{got}'"
        )


class ConsentRequiredInfo(ErrorInfo):
    """
    ConsentRequiredInfo objects contain required consent information for an error. They
    test as truthy if the error was marked as a ConsentRequired error.

    :ivar required_scopes: A list of scopes requested by the server
    :vartype required_scopes: list of str, optional
    """

    def __init__(self, error_data: dict[str, t.Any]) -> None:
        # data is only considered parseable if this error has the code 'ConsentRequired'
        has_code = error_data.get("code") == "ConsentRequired"
        data = error_data if has_code else {}
        self.required_scopes = self._parse_required_scopes(data)

        # but the result is only considered valid if both parts are present
        self._has_data = has_code and bool(self.required_scopes)

    def _parse_required_scopes(self, data: dict[str, t.Any]) -> list[str]:
        if guards.is_list_of(data.get("required_scopes"), str):
            return t.cast("list[str]", data["required_scopes"])
        elif isinstance(data.get("required_scope"), str):
            return [data["required_scope"]]
        return []


class ErrorInfoContainer:
    """
    This is a wrapper type which contains various error info objects for parsed error
    data. It is attached to API errors as the ``.info`` attribute.

    :ivar authorization_parameters: A parsed AuthorizationParameterInfo object
    :ivar consent_required: A parsed ConsentRequiredInfo object
    """

    def __init__(self, error_data: dict[str, t.Any] | None) -> None:
        self.authorization_parameters = AuthorizationParameterInfo(error_data or {})
        self.consent_required = ConsentRequiredInfo(error_data or {})

    def __str__(self) -> str:
        return f"{self.authorization_parameters}|{self.consent_required}"



import json
from unittest import mock

import requests


class PickleableMockResponse(mock.NonCallableMock):
    """
    Custom Mock class which implements __setstate__ and __getstate__ so that it
    can be pickled and unpickled correctly -- thus avoiding issues with the
    tests which pickle/unpickle clients and responses (and thereby, their inner
    objects).

    The only attributes which survive the pickle/unpickle process are those
    with defined treatment in the __getstate__ and __setstate__ methods.


    NOTE: It also has to set `__class__` explicitly, which can break some mock
    functionality.


    I've tried various workarounds using copyreg to put in a custom serializer
    for things with a __class__ of requests.Response which checks to see if
    type(obj) is PickleableMockResponse and *all kinds of stuff*. None of it
    seems to work, so this is the best thing I could figure out for now.
    - Stephen (2018-09-07)
    """

    def __init__(
        self, status_code, json_body=None, text=None, headers=None, *args, **kwargs
    ) -> None:
        kwargs["spec"] = requests.Response
        super().__init__(*args, **kwargs)
        self.__class__ = PickleableMockResponse

        # after mock initialization, setup various explicit attributes
        self.status_code = status_code

        self.headers = headers or {"Content-Type": "application/json"}

        self._json_body = json_body

        self.text = text or (json.dumps(json_body) if json_body else "")

    def json(self):
        if self._json_body is not None:
            return self._json_body
        else:
            raise ValueError("globus sdk mock value error")

    def __getstate__(self):
        """Custom getstate discards most of the magical mock stuff"""
        keys = ["headers", "text", "_json_body", "status_code"]
        return {k: self.__dict__[k] for k in keys}

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __reduce__(self):
        return (
            _unpickle_pickleable_mock_response,
            (self.status_code, self.__getstate__()),
        )


def _unpickle_pickleable_mock_response(status, state):
    x = PickleableMockResponse(status)
    x.__setstate__(state)
    return x



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_ID = str(uuid.uuid1())
TASK_ID = str(uuid.uuid1())


RESPONSES = ResponseSet(
    metadata={"index_id": INDEX_ID, "task_id": TASK_ID},
    default=RegisteredResponse(
        service="search",
        path=f"/v1/index/{INDEX_ID}/batch_delete_by_subject",
        method="POST",
        json={"task_id": TASK_ID},
    ),
)



import datetime

import globus_sdk
from globus_sdk.experimental.globus_app import UserApp

# Tutorial Client ID - <replace this with your own client>
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

# Globus Tutorial Collection 1
# https://app.globus.org/file-manager/collections/6c54cade-bde5-45c1-bdea-f4bd71dba2cc
SRC_COLLECTION = "6c54cade-bde5-45c1-bdea-f4bd71dba2cc"
SRC_PATH = "/share/godata/file1.txt"

# Globus Tutorial Collection 2
# https://app.globus.org/file-manager/collections/31ce9ba0-176d-45a5-add3-f37d233ba47d
DST_COLLECTION = "31ce9ba0-176d-45a5-add3-f37d233ba47d"
DST_PATH = "/~/example-timer-destination.txt"

# as with an immediate data transfer, we take our input data and wrap them in
# a TransferData object, representing the transfer task
transfer_request = globus_sdk.TransferData(SRC_COLLECTION, DST_COLLECTION)
transfer_request.add_item(SRC_PATH, DST_PATH)

# we'll define the timer as one which runs every hour for 3 days
# declare these data in the form of a "schedule" for the timer
#
# a wide variety of schedules are possible here; to setup a recurring timer:
# - you MUST declare an interval for the timer (`interval_seconds`)
# - you MAY declare an end condition (`end`)
schedule = globus_sdk.RecurringTimerSchedule(
    interval_seconds=3600,
    end={
        "condition": "time",
        "datetime": datetime.datetime.now() + datetime.timedelta(days=3),
    },
)

with UserApp("manage-timers-example", client_id=NATIVE_CLIENT_ID) as app:
    # create a TimersClient to interact with the service, and register any data_access
    # scopes for the collections
    with globus_sdk.TimersClient(app=app) as timers_client:
        # Omit this step if the collections are either
        #   (1) A guest collection or (2) high assurance.
        timers_client.add_app_transfer_data_access_scope(
            (SRC_COLLECTION, DST_COLLECTION)
        )

        # submit the creation request to the service, printing out the
        # ID of your new timer after it's created -- you can find it in
        # https://app.globus.org/activity/timers
        timer = timers_client.create_timer(
            globus_sdk.TransferTimer(
                name=(
                    "create-timer-example "
                    f"[created at {datetime.datetime.now().isoformat()}]"
                ),
                body=transfer_request,
                schedule=schedule,
            )
        )

    print("Finished submitting timer.")
    print(f"timer_id: {timer['timer']['job_id']}")



# check IdentityMap usages
import globus_sdk

# create clients for later usage
ac = globus_sdk.AuthClient()
nc = globus_sdk.NativeAppAuthClient("foo_client_id")
cc = globus_sdk.ConfidentialAppAuthClient("foo_client_id", "foo_client_secret")

# check init allows the service client but not the login clients
im = globus_sdk.IdentityMap(ac)
im = globus_sdk.IdentityMap(cc)  # type: ignore[arg-type]
im = globus_sdk.IdentityMap(nc)  # type: ignore[arg-type]

# getitem and delitem work, but setitem and contains do not
foo = im["foo"]
del im["foo"]
im["foo"] = "bar"  # type: ignore[index]
somebool = "foo" in im  # type: ignore[operator]



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TIMER_ID

RESPONSES = ResponseSet(
    metadata={"job_id": TIMER_ID},
    default=RegisteredResponse(
        service="timer",
        path=f"/jobs/{TIMER_ID}/pause",
        method="POST",
        json={"message": f"Successfully paused job {TIMER_ID}."},
    ),
)



from __future__ import annotations

import typing as t

from globus_sdk._internal import guards
from globus_sdk.exc import ErrorSubdocument, GlobusAPIError


class TimersAPIError(GlobusAPIError):
    """
    Error class to represent error responses from Timers.

    Implements a dedicated method for parsing error responses from Timers due
    to the differences between various error formats used.
    """

    def _parse_undefined_error_format(self) -> bool:
        """
        Treat any top-level "error" key as an "array of size 1".
        Meaning that we'll see a single subdocument for data shaped like
            {
              "error": {
                "foo": "bar"
              }
            }

        Error shapes also include validation errors in a 'details' array:

            {
                "detail": [
                    {
                        "loc": ["body", "start"],
                        "msg": "field required",
                        "type": "value_error.missing"
                    },
                    {
                        "loc": ["body", "callback_url"],
                        "msg": "field required",
                        "type": "value_error.missing"
                    }
                ]
            }
        """
        # if there is not a top-level 'error' key and no top-level
        # 'detail' key, no special behavior is defined
        # fall-back to the base class implementation
        # but before that fallback, try the two relevant branches

        # if 'error' is present, use it to populate the errors array
        # extract 'code' and 'messages' from it
        if isinstance(self._dict_data.get("error"), dict):
            self.errors = [ErrorSubdocument(self._dict_data["error"])]
            self.code = self._extract_code_from_error_array(self.errors)
            self.messages = self._extract_messages_from_error_array(self.errors)
            return True
        elif guards.is_list_of(self._dict_data.get("detail"), dict):
            # collect the errors array from details
            self.errors = [
                ErrorSubdocument(d, message_fields=("msg",))
                for d in self._dict_data["detail"]
            ]
            # extract a 'code' if there is one
            self.code = self._extract_code_from_error_array(self.errors)

            # build custom 'messages' for this case
            self.messages = [
                f"{message}: {loc}"
                for (message, loc) in _parse_detail_docs(self.errors)
            ]
            return True
        else:
            return super()._parse_undefined_error_format()


def _parse_detail_docs(
    errors: list[ErrorSubdocument],
) -> t.Iterator[tuple[str, str]]:
    for d in errors:
        if d.message is None:
            continue
        loc_list = d.get("loc")
        if not guards.is_list_of(loc_list, str):
            continue
        yield (d.message, ".".join(loc_list))



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

source_id = str(uuid.uuid4())
destination_id = str(uuid.uuid4())
owner_id = str(uuid.uuid4())
task_id = str(uuid.uuid4())

TASK_LIST_DOC = {
    "DATA": [
        {
            "bytes_checksummed": 0,
            "bytes_transferred": 4,
            "canceled_by_admin": None,
            "canceled_by_admin_message": None,
            "command": "API 0.10",
            "completion_time": "2021-09-02T18:04:49+00:00",
            "deadline": "2021-09-03T18:04:47+00:00",
            "delete_destination_extra": False,
            "destination_endpoint": "mydest",
            "destination_endpoint_display_name": "Destination Endpoint",
            "destination_endpoint_id": destination_id,
            "directories": 0,
            "effective_bytes_per_second": 3,
            "encrypt_data": False,
            "fail_on_quota_errors": False,
            "fatal_error": None,
            "faults": 0,
            "files": 1,
            "files_skipped": 0,
            "files_transferred": 1,
            "filter_rules": None,
            "history_deleted": False,
            "is_ok": None,
            "is_paused": False,
            "label": None,
            "nice_status": None,
            "nice_status_details": None,
            "nice_status_expires_in": None,
            "nice_status_short_description": None,
            "owner_id": owner_id,
            "preserve_timestamp": False,
            "recursive_symlinks": "ignore",
            "request_time": "2021-09-02T18:04:47+00:00",
            "skip_source_errors": False,
            "source_endpoint": "mysrc",
            "source_endpoint_display_name": "Source Endpoint",
            "source_endpoint_id": source_id,
            "status": "SUCCEEDED",
            "subtasks_canceled": 0,
            "subtasks_expired": 0,
            "subtasks_failed": 0,
            "subtasks_pending": 0,
            "subtasks_retrying": 0,
            "subtasks_skipped_errors": 0,
            "subtasks_succeeded": 2,
            "subtasks_total": 2,
            "symlinks": 0,
            "sync_level": None,
            "task_id": task_id,
            "type": "TRANSFER",
            "username": "u_XrtivK6z9w2MZwr65os6nZX0wv",
            "verify_checksum": True,
        }
    ],
    "DATA_TYPE": "task_list",
    "length": 1,
    "limit": 1000,
    "offset": 0,
    "total": 1,
}


RESPONSES = ResponseSet(
    metadata={
        "tasks": [
            {
                "source_id": source_id,
                "destination_id": destination_id,
                "owner_id": owner_id,
                "task_id": task_id,
            }
        ],
    },
    default=RegisteredResponse(
        service="transfer",
        path="/v0.10/task_list",
        json=TASK_LIST_DOC,
    ),
)



import pytest

from globus_sdk.authorizers import AccessTokenAuthorizer

TOKEN = "DUMMY_TOKEN"


@pytest.fixture
def authorizer():
    return AccessTokenAuthorizer(TOKEN)


def test_get_authorization_header(authorizer):
    """
    Get authorization header, confirms expected value
    """
    assert authorizer.get_authorization_header() == "Bearer " + TOKEN


def test_handle_missing_authorization(authorizer):
    """
    Confirms that AccessTokenAuthorizer doesn't handle missing authorization
    """
    assert not authorizer.handle_missing_authorization()



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

BOOKMARK_ID = str(uuid.uuid4())

_collection_id = str(uuid.uuid4())
_name = "private datasets"
_path = "/data_repository/private"
_pinned = True


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="PATCH",
        path=f"/v2/bookmarks/{BOOKMARK_ID}",
        status=200,
        json={
            "meta": {"request_id": "dDaKboC8I"},
            "data": {
                "type": "Bookmark",
                "id": BOOKMARK_ID,
                "attributes": {
                    "name": _name,
                    "path": _path,
                    "pinned": _pinned,
                },
                "relationships": {
                    "collection": {
                        "data": {
                            "type": "Collection",
                            "id": _collection_id,
                        }
                    }
                },
            },
        },
        metadata={
            "bookmark_id": BOOKMARK_ID,
            "collection": _collection_id,
            "name": _name,
            "path": _path,
            "pinned": _pinned,
        },
    )
)



from __future__ import annotations

import typing as t

from docutils.parsers.rst import directives

from ..utils import derive_doc_url_base
from .add_content_directive import AddContentDirective


class ExternalDocLink(AddContentDirective):
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    # allow for spaces in the argument string
    final_argument_whitespace = True
    option_spec = {
        "base_url": directives.unchanged,
        "service": directives.unchanged,
        "ref": directives.unchanged_required,
    }

    def gen_rst(self) -> t.Iterator[str]:
        message = self.arguments[0].strip()

        service: str | None = self.options.get("service")
        default_base_url: str = derive_doc_url_base(service)

        base_url = self.options.get("base_url", default_base_url)
        relative_link = self.options["ref"]
        # use a trailing `__` to make the hyperlink an "anonymous hyperlink"
        #
        # the reason for this is that sphinx will warn (error with -W) if we generate
        # rst with duplicate target names, as when an autodoc method name matches the
        # (snake_cased) message for a hyperlink, a common scenario
        # the conflicts are stipulated by docutils to cause warnings, along with a
        # conflict resolution procedure specified under the implicit hyperlink section
        # of the spec:
        # https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#implicit-hyperlink-targets
        #
        # for details on anonymous hyperlinks, see also:
        # https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#anonymous-hyperlinks
        yield (
            f"See `{message} <{base_url}/{relative_link}>`__ in the "
            "API documentation for details."
        )



from __future__ import annotations

import pytest

from globus_sdk import FlowsAPIError, SpecificFlowClient
from globus_sdk.testing import load_response


def test_validate_run(specific_flow_client_class: type[SpecificFlowClient]):
    metadata = load_response(SpecificFlowClient.validate_run).metadata

    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])

    resp = flow_client.validate_run(body=metadata["request_body"])
    assert resp.http_status == 200


def test_validate_run_returns_error_for_invalid_payload(
    specific_flow_client_class: type[SpecificFlowClient],
):
    metadata = load_response(
        SpecificFlowClient.validate_run, case="invalid_input_payload"
    ).metadata

    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])

    with pytest.raises(FlowsAPIError) as error:
        flow_client.validate_run(body=metadata["request_body"])
    assert error.value.http_status == 400


def test_validate_run_returns_error_for_lacking_run_permission(
    specific_flow_client_class: type[SpecificFlowClient],
):
    metadata = load_response(
        SpecificFlowClient.validate_run, case="not_a_flow_starter"
    ).metadata

    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])

    with pytest.raises(FlowsAPIError) as error:
        flow_client.validate_run(body=metadata["request_body"])
    assert error.value.http_status == 403



from responses import matchers

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

VALIDATE_SIMPLE_FLOW_DEFINITION = {
    "Comment": "Simple flow",
    "StartAt": "Step1",
    "States": {
        "Step1": {
            "Type": "Action",
            "ActionUrl": "https://transfer.actions.globus.org/transfer",
            "Parameters": {
                "source_endpoint.$": "$.source_endpoint_id",
                "destination_endpoint.$": "$.destination_endpoint_id",
                "DATA": [
                    {
                        "source_path.$": "$.source_path",
                        "destination_path.$": "$.destination_path",
                    }
                ],
            },
            "ResultPath": "$.TransferResult",
            "End": True,
        }
    },
}

VALIDATE_SIMPLE_SUCCESS_RESPONSE = {
    "scopes": {"User": ["urn:globus:auth:scope:transfer.api.globus.org:all"]}
}

VALIDATE_INVALID_FLOW_DEFINITION = {
    "Comment": "Simple flow",
    "StartAt": "Step1",
    "States": {
        "Step1": {
            "Type": "Action",
            "ActionUrl": "https://transfer.actions.globus.org/transfer",
            "Parameters": {
                "source_endpoint.$": "$.source_endpoint_id",
                "destination_endpoint.$": "$.destination_endpoint_id",
                "DATA": [
                    {
                        "source_path.$": "$.source_path",
                        "destination_path.$": "$.destination_path",
                    }
                ],
            },
            "ResultPath": "$.TransferResult",
        }
    },
}

VALIDATE_INVALID_RESPONSE = {
    "error": {
        "code": "UNPROCESSABLE_ENTITY",
        "detail": [
            {
                "loc": ["definition", "States", "Step1"],
                "msg": (
                    "A state of type 'Action' must be defined as either terminal "
                    '("End": true) or transitional ("Next": "NextStateId")'
                ),
                "type": "value_error",
            }
        ],
        "message": (
            "1 validation error in body. $.definition.States.Step1: A state of "
            "type 'Action' must be defined as either terminal (\"End\": true) "
            'or transitional ("Next": "NextStateId")'
        ),
    },
    "debug_id": "41267e70-6788-4316-8b67-df7160166466",
}

_validate_simple_flow_request = {
    "definition": VALIDATE_SIMPLE_FLOW_DEFINITION,
}

_validate_invalid_flow_request = {
    "definition": VALIDATE_INVALID_FLOW_DEFINITION,
}

RESPONSES = ResponseSet(
    metadata={
        "success": VALIDATE_SIMPLE_FLOW_DEFINITION,
        "invalid": VALIDATE_INVALID_FLOW_DEFINITION,
    },
    default=RegisteredResponse(
        service="flows",
        path="/flows/validate",
        method="POST",
        status=200,
        json=VALIDATE_SIMPLE_SUCCESS_RESPONSE,
        match=[
            matchers.json_params_matcher(
                params={"definition": VALIDATE_SIMPLE_FLOW_DEFINITION},
                strict_match=False,
            )
        ],
    ),
    definition_error=RegisteredResponse(
        service="flows",
        path="/flows/validate",
        method="POST",
        status=422,
        json=VALIDATE_INVALID_RESPONSE,
        match=[
            matchers.json_params_matcher(
                params={"definition": VALIDATE_INVALID_FLOW_DEFINITION},
                strict_match=False,
            )
        ],
    ),
)



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import ENDPOINT_ID as SRC_ENDPOINT_ID
from ._common import TASK_ID

DEST_ENDPOINT_ID = str(uuid.uuid4())
OWNER_ID = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="GET",
        path="/v0.10/endpoint_manager/task_list",
        metadata={
            "task_id": TASK_ID,
            "source": SRC_ENDPOINT_ID,
            "destination": DEST_ENDPOINT_ID,
            "owner_id": OWNER_ID,
        },
        json={
            "DATA": [
                {
                    "DATA_TYPE": "task",
                    "bytes_checksummed": 0,
                    "bytes_transferred": 14,
                    "canceled_by_admin": None,
                    "canceled_by_admin_message": None,
                    "command": "API 0.10",
                    "completion_time": "2024-02-06T06:59:02+00:00",
                    "deadline": "2024-02-07T06:58:59+00:00",
                    "delete_destination_extra": False,
                    "destination_base_path": None,
                    "destination_endpoint": f"pliny_the_elder#{DEST_ENDPOINT_ID}",
                    "destination_endpoint_display_name": "Ercolano",
                    "destination_endpoint_id": DEST_ENDPOINT_ID,
                    "destination_host_endpoint": None,
                    "destination_host_endpoint_display_name": None,
                    "destination_host_endpoint_id": None,
                    "destination_host_path": None,
                    "destination_local_user": "pliny",
                    "destination_local_user_status": "OK",
                    "destination_mapped_collection_display_name": None,
                    "destination_mapped_collection_id": None,
                    "directories": 2,
                    "effective_bytes_per_second": 5,
                    "encrypt_data": False,
                    "fail_on_quota_errors": False,
                    "fatal_error": None,
                    "faults": 0,
                    "files": 3,
                    "files_skipped": 0,
                    "files_transferred": 3,
                    "filter_rules": None,
                    "history_deleted": False,
                    "is_ok": None,
                    "is_paused": False,
                    "label": None,
                    "nice_status": None,
                    "nice_status_details": None,
                    "nice_status_expires_in": None,
                    "nice_status_short_description": None,
                    "owner_id": OWNER_ID,
                    "owner_string": "pliny-the-elder@globus.org",
                    "preserve_timestamp": False,
                    "recursive_symlinks": "ignore",
                    "request_time": "2024-02-06T06:58:59+00:00",
                    "skip_source_errors": False,
                    "source_base_path": None,
                    "source_endpoint": f"pliny#{SRC_ENDPOINT_ID}",
                    "source_endpoint_display_name": "Pompeii",
                    "source_endpoint_id": SRC_ENDPOINT_ID,
                    "source_host_endpoint": None,
                    "source_host_endpoint_display_name": None,
                    "source_host_endpoint_id": None,
                    "source_host_path": None,
                    "source_local_user": None,
                    "source_local_user_status": "NO_PERMISSION",
                    "source_mapped_collection_display_name": None,
                    "source_mapped_collection_id": None,
                    "status": "SUCCEEDED",
                    "subtasks_canceled": 0,
                    "subtasks_expired": 0,
                    "subtasks_failed": 0,
                    "subtasks_pending": 0,
                    "subtasks_retrying": 0,
                    "subtasks_skipped_errors": 0,
                    "subtasks_succeeded": 6,
                    "subtasks_total": 6,
                    "symlinks": 0,
                    "sync_level": None,
                    "task_id": TASK_ID,
                    "type": "TRANSFER",
                    "username": "pliny",
                    "verify_checksum": True,
                }
            ],
            "DATA_TYPE": "task_list",
            "has_next_page": False,
            "last_key": "complete,2024-02-06T06:59:02.291996",
        },
    ),
)



import pytest
import responses


def test_cannot_encode_dict_as_text(client):
    with pytest.raises(TypeError):
        client.post("/bar", data={"baz": 1}, encoding="text")


def test_cannot_encode_with_unknown_encoding(client):
    with pytest.raises(ValueError):
        client.post("/bar", data={"baz": 1}, encoding="some-random-string")


def test_cannot_form_encode_bad_types(client):
    with pytest.raises(TypeError):
        client.post("/bar", data=["baz", "buzz"], encoding="form")
    with pytest.raises(TypeError):
        client.post("/bar", data=1, encoding="form")


def test_form_encoding_works(client):
    responses.add(responses.POST, "https://foo.api.globus.org/bar", body="hi")
    client.post("/bar", data={"baz": 1}, encoding="form")

    last_req = responses.calls[-1].request
    assert last_req.body == "baz=1"


def test_text_encoding_sends_ascii_string(client):
    responses.add(responses.POST, "https://foo.api.globus.org/bar", body="hi")
    client.post("/bar", data="baz", encoding="text")

    last_req = responses.calls[-1].request
    assert last_req.body == "baz"


def test_text_encoding_can_send_non_ascii_utf8_bytes(client):
    # this test is a reproducer for an issue in which attempting to send these bytes
    # in the form of a (decoded) string would fail, as urllib3 tried to encode them as
    # latin-1 bytes incorrectly
    # passing the bytes already UTF-8 encoded should work
    responses.add(responses.POST, "https://foo.api.globus.org/bar", body="hi")
    client.post("/bar", data='{"field“: "value“}'.encode(), encoding="text")

    last_req = responses.calls[-1].request
    assert last_req.body == '{"field“: "value“}'.encode()



from globus_sdk import (
    BatchMembershipActions,
    GroupMemberVisibility,
    GroupPolicies,
    GroupRole,
    GroupVisibility,
)

act = BatchMembershipActions()

act.add_members(["foo"], role=GroupRole.member)
act.add_members(["foo"], role=GroupRole.manager)
act.add_members(["foo"], role=GroupRole.admin)
act.add_members(["foo"], role="member")
act.add_members(["foo"], role="manager")
act.add_members(["foo"], role="admin")
act.add_members(["foo"], role="unknown")  # type: ignore[arg-type]

GroupPolicies(
    is_high_assurance=False,
    group_members_visibility="members",
    join_requests=True,
    signup_fields=[],
    group_visibility=GroupVisibility.authenticated,
)
GroupPolicies(
    is_high_assurance=False,
    group_members_visibility="members",
    join_requests=True,
    signup_fields=[],
    group_visibility=GroupVisibility.private,
)
GroupPolicies(
    is_high_assurance=False,
    group_members_visibility="members",
    join_requests=True,
    signup_fields=[],
    group_visibility="authenticated",
)
GroupPolicies(
    is_high_assurance=False,
    group_members_visibility="members",
    join_requests=True,
    signup_fields=[],
    group_visibility="private",
)
GroupPolicies(
    is_high_assurance=False,
    group_members_visibility="members",
    join_requests=True,
    signup_fields=[],
    group_visibility="unknown",  # type: ignore[arg-type]
)

GroupPolicies(
    is_high_assurance=False,
    group_visibility="private",
    join_requests=True,
    signup_fields=[],
    group_members_visibility=GroupMemberVisibility.members,
)
GroupPolicies(
    is_high_assurance=False,
    group_visibility="private",
    join_requests=True,
    signup_fields=[],
    group_members_visibility=GroupMemberVisibility.managers,
)
GroupPolicies(
    is_high_assurance=False,
    group_visibility="private",
    join_requests=True,
    signup_fields=[],
    group_members_visibility="members",
)
GroupPolicies(
    is_high_assurance=False,
    group_visibility="private",
    join_requests=True,
    signup_fields=[],
    group_members_visibility="managers",
)
GroupPolicies(
    is_high_assurance=False,
    group_visibility="private",
    join_requests=True,
    signup_fields=[],
    group_members_visibility="unknown",  # type: ignore[arg-type]
)



import globus_sdk
from globus_sdk.testing import load_response


def test_lock_endpoint(compute_client_v3: globus_sdk.ComputeClientV3):
    meta = load_response(compute_client_v3.lock_endpoint).metadata

    res = compute_client_v3.lock_endpoint(endpoint_id=meta["endpoint_id"])

    assert res.http_status == 200
    assert res.data["endpoint_id"] == meta["endpoint_id"]



"""
Unit tests for globus_sdk.SearchQueryV1
"""

from globus_sdk import MISSING, SearchQueryV1


def test_init_v1():
    query = SearchQueryV1()

    # ensure the version is set to query#1.0.0
    assert query["@version"] == "query#1.0.0"

    # ensure key attributes initialize to empty lists
    for attribute in ["facets", "filters", "post_facet_filters", "sort", "boosts"]:
        assert query[attribute] == MISSING

    # init with supported fields
    params = {"q": "foo", "limit": 10, "offset": 0, "advanced": False}
    param_query = SearchQueryV1(**params)
    for par in params:
        assert param_query[par] == params[par]

    # init with additional_fields
    add_params = {"param1": "value1", "param2": "value2"}
    param_query = SearchQueryV1(additional_fields=add_params)
    for par in add_params:
        assert param_query[par] == add_params[par]



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_ID = str(uuid.uuid4())

_default_display_name = "Awesome Index of Awesomeness"
_default_description = "An index so awesome that it simply cannot be described"

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="search",
        method="PATCH",
        path=f"/v1/index/{INDEX_ID}",
        json={
            "@datatype": "GSearchIndex",
            "@version": "2017-09-01",
            "creation_date": "2021-04-05 15:05:18",
            "display_name": _default_display_name,
            "description": _default_description,
            "id": INDEX_ID,
            "is_trial": True,
            "subscription_id": None,
            "max_size_in_mb": 1,
            "num_entries": 0,
            "num_subjects": 0,
            "size_in_mb": 0,
            "status": "open",
        },
        metadata={"index_id": INDEX_ID, "display_name": _default_display_name},
    ),
    forbidden=RegisteredResponse(
        service="search",
        method="PATCH",
        path=f"/v1/index/{INDEX_ID}",
        status=403,
        json={
            "@datatype": "GError",
            "@version": "2017-09-01",
            "status": 403,
            "code": "Forbidden.Generic",
            "message": "index_update request denied by service",
            "request_id": "0e73b6a61e53468684f86c7993336a72",
            "error_data": {
                "cause": (
                    "You do not have the proper roles "
                    "to perform the index_update operation."
                ),
                "recommended_resolution": (
                    "Ensure you are making a call authenticated with "
                    "a valid Search token and that you have been granted "
                    "the required roles for this operation"
                ),
            },
        },
        metadata={"index_id": INDEX_ID},
    ),
)



from __future__ import annotations

import enum
import typing as t

import requests

if t.TYPE_CHECKING:
    from .caller_info import RequestCallerInfo

C = t.TypeVar("C", bound=t.Callable[..., t.Any])


# alias useful for declaring retry-related types
RetryCheck = t.Callable[["RetryContext"], "RetryCheckResult"]


class RetryContext:
    """
    The RetryContext is an object passed to retry checks in order to determine whether
    or not a request should be retried. The context is constructed after each request,
    regardless of success or failure.

    If an exception was raised, the context will contain that exception object.
    Otherwise, the context will contain a response object. Exactly one of ``response``
    or ``exception`` will be present.

    :param attempt: The request attempt number, starting at 0.
    :param caller_info: Contextual information about the caller, including authorizer
    :param response: The response on a successful request
    :param exception: The error raised when trying to send the request
    """

    def __init__(
        self,
        attempt: int,
        *,
        caller_info: RequestCallerInfo,
        response: requests.Response | None = None,
        exception: Exception | None = None,
    ) -> None:
        # retry attempt number
        self.attempt = attempt
        # caller info provides contextual information about the request
        self.caller_info = caller_info
        # the response or exception from a request
        # we expect exactly one of these to be non-null
        self.response = response
        self.exception = exception
        # the retry delay or "backoff" before retrying
        self.backoff: float | None = None


class RetryCheckResult(enum.Enum):
    #: yes, retry the request
    do_retry = enum.auto()
    #: no, do not retry the request
    do_not_retry = enum.auto()
    #: "I don't know", ask other checks for an answer
    no_decision = enum.auto()


class RetryCheckFlags(enum.Flag):
    #: no flags (default)
    NONE = enum.auto()
    #: only run this check once per request
    RUN_ONCE = enum.auto()


# stub for mypy
class _RetryCheckFunc:
    _retry_check_flags: RetryCheckFlags


def set_retry_check_flags(flag: RetryCheckFlags) -> t.Callable[[C], C]:
    """
    A decorator for setting retry check flags on a retry check function.
    Usage:

    >>> @set_retry_check_flags(RetryCheckFlags.RUN_ONCE)
    >>> def foo(ctx): ...

    :param flag: The flag to set on the check
    """

    def decorator(func: C) -> C:
        as_check = t.cast(_RetryCheckFunc, func)
        as_check._retry_check_flags = flag
        return func

    return decorator


class RetryCheckCollection:
    """
    A RetryCheckCollection is an ordered collection of retry checks which are
    used to determine whether or not a request should be retried.

    Checks are stored in registration order.

    Notably, the collection does not decide
    - how many times a request should retry
    - how or how long the call should wait between attempts
      (except via the backoff which may be set)
    - what kinds of request parameters (e.g., timeouts) are used

    It *only* contains ``RetryCheck`` functions which can look at a response or
    error and decide whether or not to retry.
    """

    def __init__(self) -> None:
        self._data: list[RetryCheck] = []

    def register_check(self, func: RetryCheck) -> RetryCheck:
        """
        Register a retry check with this policy.

        A retry checker is a callable responsible for implementing
        `check(RetryContext) -> RetryCheckResult`

        `check` should *not* perform any sleeps or delays.
        Multiple checks should be chainable and callable in any order.

        :param func: The function or other callable to register as a retry check
        """
        self._data.append(func)
        return func

    def register_many_checks(self, funcs: t.Iterable[RetryCheck]) -> None:
        """
        Register all checks in a collection of checks.

        :param funcs: An iterable collection of retry check callables
        """
        for f in funcs:
            self.register_check(f)

    def __iter__(self) -> t.Iterator[RetryCheck]:
        yield from self._data

    def __len__(self) -> int:
        return len(self._data)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

globusid_idp = {
    "short_name": "globusid",
    "id": "41143743-f3c8-4d60-bbdb-eeecaba85bd9",
    "alternative_names": [],
    "domains": ["globusid.org"],
    "name": "Globus ID",
}

globus_staff_idp = {
    "short_name": "globus.org",
    "id": "927d7238-f917-4eb2-9ace-c523fa9ba34e",
    "alternative_names": [],
    "domains": ["globus.org"],
    "name": "Globus Staff",
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path="/v2/api/identity_providers",
        json={"identity_providers": [globusid_idp, globus_staff_idp]},
        metadata={
            "ids": [globusid_idp["id"], globus_staff_idp["id"]],
            "domains": [globusid_idp["domains"][0], globus_staff_idp["domains"][0]],
        },
    ),
    globusid=RegisteredResponse(
        service="auth",
        path="/v2/api/identity_providers",
        json={"identity_providers": [globusid_idp]},
        metadata={"id": globusid_idp["id"], "domains": globusid_idp["domains"]},
    ),
    globus_staff=RegisteredResponse(
        service="auth",
        path="/v2/api/identity_providers",
        json={"identity_providers": [globus_staff_idp]},
        metadata={"id": globus_staff_idp["id"], "domains": globus_staff_idp["domains"]},
    ),
)



from __future__ import annotations

import dataclasses
import sys
import typing as t

# pass slots=True on 3.10+
# it's not strictly necessary, but it improves performance
if sys.version_info >= (3, 10):
    _add_dataclass_kwargs: dict[str, bool] = {"slots": True}
else:
    _add_dataclass_kwargs: dict[str, bool] = {}


@dataclasses.dataclass(frozen=True, repr=False, **_add_dataclass_kwargs)
class Scope:
    """
    A scope object is a representation of a scope and its dynamic dependencies
    (other scopes).

    A scope may be optional (also referred to as "atomically revocable").
    An optional scope can be revoked without revoking consent for other scopes
    which were granted at the same time.

    Scopes are immutable, and provide several evolver methods which produce new
    Scopes. In particular, ``with_dependency`` and ``with_dependencies`` create
    new scopes with added dependencies.

    ``str(Scope(...))`` produces a valid scope string for use in various methods.

    :param scope_string: The string which will be used as the basis for this Scope
    :param optional: The scope may be marked as optional. This means that the scope can
        be declined by the user without declining consent for other scopes.
    """

    scope_string: str
    optional: bool = dataclasses.field(default=False)
    dependencies: tuple[Scope, ...] = dataclasses.field(default=())

    def __post_init__(
        self,
    ) -> None:
        if any(c in self.scope_string for c in "[]* "):
            raise ValueError(
                "Scope instances may not contain the special characters '[]* '. "
                "Use Scope.parse instead."
            )

    @classmethod
    def parse(cls, scope_string: str) -> Scope:
        """
        Deserialize a scope string to a scope object.

        This is the special case of parsing in which exactly one scope must be returned
        by the parse. If more than one scope is returned by the parse, a ``ValueError``
        will be raised.

        :param scope_string: The string to parse
        """
        # deferred import because ScopeParser depends on Scope, but Scope.parse
        # is a wrapper over ScopeParser.parse()
        from .parser import ScopeParser

        data = ScopeParser.parse(scope_string)
        if len(data) != 1:
            raise ValueError(
                "`Scope.parse()` did not get exactly one scope. "
                f"Instead got data={data}"
            )
        return data[0]

    def with_dependency(self, other_scope: Scope) -> Scope:
        """
        Create a new scope with a dependency.
        The dependent scope relationship will be stored in the Scope and will
        be evident in its string representation.

        :param other_scope: The scope upon which the current scope depends.
        """
        if not isinstance(other_scope, Scope):
            raise TypeError(
                "Scope.with_dependency() takes a Scope as its input. "
                f"Got: '{type(other_scope).__qualname__}'"
            )
        return dataclasses.replace(
            self, dependencies=self.dependencies + (other_scope,)
        )

    def with_dependencies(self, other_scopes: t.Iterable[Scope]) -> Scope:
        """
        Create a new scope with added dependencies.
        The dependent scope relationships will be stored in the Scope and will
        be evident in its string representation.

        :param other_scopes: The scopes upon which the current scope depends.
        """
        other_scopes_tuple = tuple(other_scopes)
        for i, item in enumerate(other_scopes_tuple):
            if not isinstance(item, Scope):
                raise TypeError(
                    "Scope.with_dependencies() takes "
                    "an iterable of Scopes as its input. "
                    f"At position {i}, got: '{type(item).__qualname__}'"
                )
        return dataclasses.replace(
            self, dependencies=self.dependencies + other_scopes_tuple
        )

    def with_optional(self, optional: bool) -> Scope:
        """
        Create a new scope with a different 'optional' value.

        :param optional: Whether or not the scope is optional.
        """
        return dataclasses.replace(self, optional=optional)

    def __repr__(self) -> str:
        parts: list[str] = [f"'{self.scope_string}'"]
        if self.optional:
            parts.append("optional=True")
        if self.dependencies:
            parts.append(f"dependencies={self.dependencies!r}")
        return "Scope(" + ", ".join(parts) + ")"

    def __str__(self) -> str:
        base_scope = ("*" if self.optional else "") + self.scope_string
        if not self.dependencies:
            return base_scope
        return base_scope + "[" + " ".join(str(c) for c in self.dependencies) + "]"



from __future__ import annotations

import typing as t

import sphinx.application


def after_autodoc_signature_replace_MISSING_repr(  # pylint: disable=missing-param-doc,missing-type-doc  # noqa: E501
    app: sphinx.application.Sphinx,  # pylint: disable=unused-argument
    what: str,  # pylint: disable=unused-argument
    name: str,  # pylint: disable=unused-argument
    obj: object,  # pylint: disable=unused-argument
    options: t.Any,  # pylint: disable=unused-argument
    signature: str | None,
    return_annotation: str | None,
) -> tuple[str | None, str | None]:
    """
    convert <globus_sdk.MISSING> to MISSING in autodoc signatures

    :param signature: the signature after autodoc parsing/rendering
    :param return_annotation: the return type annotation, including the leading `->`,
        after autodoc parsing/rendering
    """
    if signature is not None:
        signature = signature.replace("<globus_sdk.MISSING>", "MISSING")
    if return_annotation is not None:
        return_annotation = return_annotation.replace("<globus_sdk.MISSING>", "MISSING")
    return signature, return_annotation



import json
import os

import pytest

from globus_sdk import __version__
from globus_sdk.token_storage import JSONTokenStorage
from globus_sdk.token_storage.legacy import SimpleJSONFileAdapter

IS_WINDOWS = os.name == "nt"


@pytest.fixture
def json_file(tmp_path):
    return tmp_path / "mydata.json"


def test_file_does_not_exist(json_file):
    adapter = JSONTokenStorage(json_file)
    assert not adapter.file_exists()


def test_file_exists(json_file):
    json_file.touch()
    adapter = JSONTokenStorage(json_file)
    assert adapter.file_exists()


def test_store_and_get_token_data_by_resource_server(
    json_file, mock_token_data_by_resource_server
):
    adapter = JSONTokenStorage(json_file)
    assert not adapter.file_exists()
    adapter.store_token_data_by_resource_server(mock_token_data_by_resource_server)

    gotten = adapter.get_token_data_by_resource_server()

    for resource_server in ["resource_server_1", "resource_server_2"]:
        assert (
            mock_token_data_by_resource_server[resource_server].to_dict()
            == gotten[resource_server].to_dict()
        )


def test_store_token_response_with_namespace(json_file, mock_response):
    adapter = JSONTokenStorage(json_file, namespace="foo")
    assert not adapter.file_exists()
    adapter.store_token_response(mock_response)

    data = json.loads(json_file.read_text())
    assert data["globus-sdk.version"] == __version__
    assert data["data"]["foo"]["resource_server_1"]["access_token"] == "access_token_1"
    assert data["data"]["foo"]["resource_server_2"]["access_token"] == "access_token_2"


def test_get_token_data(json_file, mock_response):
    adapter = JSONTokenStorage(json_file)
    assert not adapter.file_exists()
    adapter.store_token_response(mock_response)

    assert adapter.get_token_data("resource_server_1").access_token == "access_token_1"


def test_remove_token_data(json_file, mock_response):
    adapter = JSONTokenStorage(json_file)
    assert not adapter.file_exists()
    adapter.store_token_response(mock_response)

    # remove rs1, confirm only rs2 is still available
    remove_result = adapter.remove_token_data("resource_server_1")
    assert remove_result is True

    assert adapter.get_token_data("resource_server_1") is None
    assert adapter.get_token_data("resource_server_2").access_token == "access_token_2"

    # confirm unable to re-remove rs1
    remove_result = adapter.remove_token_data("resource_server_1")
    assert remove_result is False


@pytest.mark.xfail(IS_WINDOWS, reason="cannot set umask perms on Windows")
def test_store_perms(json_file, mock_response):
    adapter = JSONTokenStorage(json_file)
    assert not adapter.file_exists()
    adapter.store_token_response(mock_response)

    # mode|0600 should be 0600 -- meaning that those are the maximal
    # permissions given
    st_mode = json_file.stat().st_mode & 0o777  # & 777 to remove extra bits
    assert st_mode | 0o600 == 0o600


def test_migrate_from_v1_adapter(json_file, mock_response):
    # write with a SimpleJSONFileAdapter
    old_adapter = SimpleJSONFileAdapter(json_file)
    old_adapter.store(mock_response)

    # confirm able to read with JSONTokenStorage
    new_adapter = JSONTokenStorage(json_file)
    assert (
        new_adapter.get_token_data("resource_server_1").access_token == "access_token_1"
    )

    # confirm version is overwritten on next store
    new_adapter.store_token_response(mock_response)
    data = json.loads(json_file.read_text())
    assert data["format_version"] == "2.0"



import urllib.parse

import pytest

from globus_sdk import MISSING
from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize("filter_roles", [MISSING, "owner"])
@pytest.mark.parametrize("orderby", [MISSING, "name ASC"])
def test_list_registered_apis_simple(flows_client, filter_roles, orderby):
    meta = load_response(flows_client.list_registered_apis).metadata

    add_kwargs = {}
    if filter_roles is not MISSING:
        add_kwargs["filter_roles"] = filter_roles
    if orderby is not MISSING:
        add_kwargs["orderby"] = orderby

    res = flows_client.list_registered_apis(**add_kwargs)

    assert res.http_status == 200
    # dict-like indexing
    assert meta["first_registered_api_id"] == res["registered_apis"][0]["id"]
    # list conversion (using __iter__) and indexing
    assert meta["first_registered_api_id"] == list(res)[0]["id"]

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    expect_query_params = {
        k: [v]
        for k, v in (
            ("filter_roles", filter_roles),
            ("orderby", orderby),
        )
        if v is not MISSING
    }
    assert parsed_qs == expect_query_params


@pytest.mark.parametrize("by_pages", [True, False])
def test_list_registered_apis_paginated(flows_client, by_pages):
    meta = load_response(flows_client.list_registered_apis, case="paginated").metadata
    total_items = meta["total_items"]
    num_pages = meta["num_pages"]
    expect_markers = meta["expect_markers"]

    res = flows_client.paginated.list_registered_apis()
    if by_pages:
        pages = list(res)
        assert len(pages) == num_pages
        for i, page in enumerate(pages):
            assert page["marker"] == expect_markers[i]
            if i < num_pages - 1:
                assert page["has_next_page"] is True
            else:
                assert page["has_next_page"] is False
    else:
        items = list(res.items())
        assert len(items) == total_items



import pytest

from globus_sdk.token_storage import (
    JSONTokenStorage,
    MemoryTokenStorage,
    SQLiteTokenStorage,
)


@pytest.fixture(params=["json", "sqlite", "memory"])
def storage(request, tmp_path):
    if request.param == "json":
        file = tmp_path / "mydata.json"
        yield JSONTokenStorage(file)
    elif request.param == "sqlite":
        file = tmp_path / "mydata.db"
        store = SQLiteTokenStorage(file)
        yield store
        store.close()
    else:
        yield MemoryTokenStorage()


def test_store_authorization_code_response(
    storage, authorization_code_response, id_token_sub
):
    storage.store_token_response(authorization_code_response)

    tok_by_rs = authorization_code_response.by_resource_server

    stored_data = storage.get_token_data_by_resource_server()

    for resource_server in ["resource_server_1", "resource_server_2"]:
        for fieldname in (
            "resource_server",
            "scope",
            "access_token",
            "refresh_token",
            "expires_at_seconds",
            "token_type",
        ):
            assert tok_by_rs[resource_server][fieldname] == getattr(
                stored_data[resource_server], fieldname
            )
        assert "identity_id" not in tok_by_rs[resource_server]
        assert stored_data[resource_server].identity_id == id_token_sub


def test_store_dependent_token_response(storage, dependent_token_response):
    """
    If a TokenStorage is asked to store dependent token data, it should work and
    produce identity_id values of None (because there is no id_token to inspect)
    """
    storage.store_token_response(dependent_token_response)

    dep_tok_by_rs = dependent_token_response.by_resource_server

    stored_data = storage.get_token_data_by_resource_server()

    for resource_server in ["resource_server_1", "resource_server_2"]:
        for fieldname in (
            "resource_server",
            "scope",
            "access_token",
            "refresh_token",
            "expires_at_seconds",
            "token_type",
        ):
            assert dep_tok_by_rs[resource_server][fieldname] == getattr(
                stored_data[resource_server], fieldname
            )
        assert stored_data[resource_server].identity_id is None
        assert "identity_id" not in dep_tok_by_rs[resource_server]


def test_store_refresh_token_response(storage, refresh_token_response):
    """
    If a TokenStorage is asked to store refresh token data, it should work and
    produce identity_id values of None (because there is no id_token to inspect)
    """
    storage.store_token_response(refresh_token_response)

    refresh_tok_by_rs = refresh_token_response.by_resource_server

    stored_data = storage.get_token_data_by_resource_server()

    for fieldname in (
        "resource_server",
        "scope",
        "access_token",
        "refresh_token",
        "expires_at_seconds",
        "token_type",
    ):
        assert refresh_tok_by_rs["resource_server_1"][fieldname] == getattr(
            stored_data["resource_server_1"], fieldname
        )
    assert stored_data["resource_server_1"].identity_id is None
    assert "identity_id" not in refresh_tok_by_rs["resource_server_1"]



from __future__ import annotations

import logging

from globus_sdk import exc

log = logging.getLogger(__name__)


class AuthAPIError(exc.GlobusAPIError):
    """
    Error class for the API components of Globus Auth.
    """

    def _post_parse_hook(self) -> bool:
        # if there was only one error ID set in the response, use that as the request_id
        # this allows for some errors to omit the 'id':
        #
        #    errors=[{"id": "foo"}, {}]
        #
        # or for all errors to have the same 'id':
        #
        #    errors=[{"id": "foo"}, {"id": "foo"}]
        #
        # but not for errors to have mixed/different 'id' values:
        #
        #    errors=[{"id": "foo"}, {"id": "bar"}]

        # step 1, collect error IDs
        error_ids = {suberror.get("id") for suberror in self.errors}
        # step 2, remove `None` from any sub-errors which did not set error_id or
        # explicitly set it to null
        error_ids.discard(None)
        # step 3, check if there was exactly one error ID and it was a string
        if len(error_ids) == 1:
            maybe_error_id = error_ids.pop()
            if isinstance(maybe_error_id, str):
                self.request_id = maybe_error_id

        return True



import typing as t
import uuid

from responses.matchers import json_params_matcher

from globus_sdk import DependentScopeSpec
from globus_sdk.testing.models import RegisteredResponse, ResponseSet

SCOPE_REQUEST_ARGS = {
    "client_id": str(uuid.uuid1()),
    "name": "Client manage scope",
    "description": "Manage configuration of this client",
    "scope_suffix": "manage",
}


def make_request_body(request_args: t.Mapping[str, t.Any]) -> t.Dict[str, t.Any]:
    request_body = {
        "name": request_args["name"],
        "description": request_args["description"],
        "scope_suffix": request_args["scope_suffix"],
    }

    if "advertised" in request_args:
        request_body["advertised"] = request_args["advertised"]

    if "allows_refresh_token" in request_args:
        request_body["allows_refresh_token"] = request_args["allows_refresh_token"]

    if "required_domains" in request_args:
        request_body["required_domains"] = request_args["required_domains"]

    if "dependent_scopes" in request_args:
        request_body["dependent_scopes"] = request_args["dependent_scopes"]

    return request_body


def make_response_body(request_args: t.Mapping[str, t.Any]) -> t.Dict[str, t.Any]:
    return {
        "scope_string": f"https://auth.globus.org/scopes/{request_args['client_id']}/{request_args['scope_suffix']}",  # noqa: E501
        "allows_refresh_token": request_args.get("allows_refresh_token", True),
        "id": str(uuid.uuid1()),
        "advertised": request_args.get("advertised", False),
        "required_domains": request_args.get("required_domains", []),
        "name": request_args["name"],
        "description": request_args["description"],
        "client": str(request_args["client_id"]),
        "dependent_scopes": [
            {
                "scope": str(ds["scope"]),
                "optional": ds["optional"],
                "requires_refresh_token": ds["requires_refresh_token"],
            }
            for ds in request_args.get("dependent_scopes", [])
        ],
    }


def register_response(
    args: t.Mapping[str, t.Any],
) -> RegisteredResponse:
    request_args = {**SCOPE_REQUEST_ARGS, **args}
    request_body = make_request_body(request_args)
    response_body = make_response_body(request_args)

    return RegisteredResponse(
        service="auth",
        method="POST",
        path=f"/v2/api/clients/{request_args['client_id']}/scopes",
        json={"scope": response_body},
        metadata={
            # Test functions use 'args' to form request
            "args": request_args,
            # Test functions use 'response' to verify response
            "response": response_body,
        },
        match=[json_params_matcher({"scope": request_body})],
    )


RESPONSES = ResponseSet(
    default=register_response({}),
    client_id_str=register_response({"client_id": str(uuid.uuid1())}),
    client_id_uuid=register_response({"client_id": uuid.uuid1()}),
    name=register_response({"name": str(uuid.uuid4()).replace("-", "")}),
    description=register_response({"description": str(uuid.uuid4()).replace("-", "")}),
    scope_suffix=register_response(
        {"scope_suffix": str(uuid.uuid4()).replace("-", "")}
    ),
    advertised=register_response({"advertised": True}),
    not_advertised=register_response({"advertised": False}),
    allows_refresh_token=register_response({"allows_refresh_token": True}),
    disallows_refresh_token=register_response({"allows_refresh_token": False}),
    no_required_domains=register_response({"required_domains": []}),
    required_domains=register_response(
        {"required_domains": ["globus.org", "uchicago.edu"]}
    ),
    no_dependent_scopes=register_response({"dependent_scopes": []}),
    dependent_scopes=register_response(
        {
            "dependent_scopes": [
                DependentScopeSpec(str(uuid.uuid1()), True, True),
                DependentScopeSpec(uuid.uuid1(), False, False),
            ],
        }
    ),
)



import urllib.parse
import uuid

import pytest

import globus_sdk
from globus_sdk.testing import get_last_request, load_response


@pytest.fixture
def client_id():
    return str(uuid.uuid4())


@pytest.fixture
def base_client(client_id):
    return globus_sdk.AuthLoginClient(client_id=client_id)


@pytest.fixture
def confidential_client(client_id):
    return globus_sdk.ConfidentialAppAuthClient(
        client_id=client_id, client_secret="somesecret"
    )


def test_oauth2_revoke_token_works(client_id, base_client):
    load_response(base_client.oauth2_revoke_token)
    response = base_client.oauth2_revoke_token("sometoken")
    assert response["active"] is False
    lastreq = get_last_request()
    body = lastreq.body
    assert body != ""
    parsed_body = urllib.parse.parse_qs(body)
    assert parsed_body == {"token": ["sometoken"], "client_id": [client_id]}


def test_oauth2_revoke_token_does_not_send_client_id_when_authenticated(
    client_id,
    confidential_client,
):
    load_response(confidential_client.oauth2_revoke_token)
    response = confidential_client.oauth2_revoke_token("sometoken")
    assert response["active"] is False
    lastreq = get_last_request()
    body = lastreq.body
    assert body != ""
    parsed_body = urllib.parse.parse_qs(body)
    assert parsed_body == {"token": ["sometoken"]}



from __future__ import annotations

import sys
import typing as t
import uuid

from globus_sdk import client, response
from globus_sdk._internal.remarshal import commajoin
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.scopes import GroupsScopes, Scope

from .data import BatchMembershipActions, GroupPolicies
from .errors import GroupsAPIError

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

_VALID_STATUSES_T: TypeAlias = t.Literal[
    "active",
    "declined",
    "invited",
    "left",
    "pending",
    "rejected",
    "removed",
]


class GroupsClient(client.BaseClient):
    """
    Client for the
    `Globus Groups API <https://docs.globus.org/api/groups/>`_.

    .. sdk-sphinx-copy-params:: BaseClient

    This provides a relatively low level client to public groups API endpoints.
    You may also consider looking at the GroupsManager as a simpler interface
    to more common actions.

    .. automethodlist:: globus_sdk.GroupsClient
    """

    error_class = GroupsAPIError
    service_name = "groups"
    scopes = GroupsScopes

    @property
    def default_scope_requirements(self) -> list[Scope]:
        return [GroupsScopes.view_my_groups_and_memberships]

    def get_my_groups(
        self,
        *,
        statuses: (
            _VALID_STATUSES_T | t.Iterable[_VALID_STATUSES_T] | MissingType
        ) = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.ArrayResponse:
        """
        Return a list of groups your identity belongs to.

        :param statuses:
            If provided, only groups containing memberships with the given status
            are returned.
            Valid values are ``active``, ``invited``, ``pending``, ``rejected``,
            ``removed``, ``left``, and ``declined``.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /v2/groups/my_groups``

                .. extdoclink:: Retrieve your groups and membership
                    :service: groups
                    :ref: get_my_groups_and_memberships_v2_groups_my_groups_get
        """
        query_params = {"statuses": commajoin(statuses), **(query_params or {})}
        return response.ArrayResponse(
            self.get("/v2/groups/my_groups", query_params=query_params)
        )

    def get_group(
        self,
        group_id: uuid.UUID | str,
        *,
        include: str | t.Iterable[str] | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get details about a specific group

        :param group_id: the ID of the group
        :param include: list of additional fields to include (allowed fields are
            ``memberships``, ``my_memberships``, ``policies``, ``allowed_actions``, and
            ``child_ids``)
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /v2/groups/<group_id>``

                .. extdoclink:: Get Group
                    :service: groups
                    :ref: get_group_v2_groups__group_id__get
        """
        query_params = {"include": commajoin(include), **(query_params or {})}
        return self.get(f"/v2/groups/{group_id}", query_params=query_params)

    def get_group_by_subscription_id(
        self, subscription_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Using a subscription ID, find the group which provides that subscription.

        :param subscription_id: the subscription ID of the group

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import GroupsClient

                    groups = GroupsClient(...)
                    group_id = groups.get_group_by_subscription_id(subscription_id)["group_id"]

            .. tab-item:: Example Response Data

                .. expandtestfixture:: groups.get_group_by_subscription_id

            .. tab-item:: API Info

                ``GET /v2/subscription_info/<subscription_id>``

                .. extdoclink:: Get Group by Subscription ID
                    :service: groups
                    :ref: get_group_by_subscription_id_v2_subscription_info__subscription_id__get
        """  # noqa: E501
        return self.get(f"/v2/subscription_info/{subscription_id}")

    def delete_group(
        self,
        group_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete a group.

        :param group_id: the ID of the group
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /v2/groups/<group_id>``

                .. extdoclink:: Delete a Group
                    :service: groups
                    :ref: delete_group_v2_groups__group_id__delete
        """
        return self.delete(f"/v2/groups/{group_id}", query_params=query_params)

    def create_group(
        self,
        data: dict[str, t.Any],
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Create a group.

        :param data: the group document to create
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``POST /v2/groups``

                .. extdoclink:: Create a Group
                    :service: groups
                    :ref: create_group_v2_groups_post
        """
        return self.post("/v2/groups", data=data, query_params=query_params)

    def update_group(
        self,
        group_id: uuid.UUID | str,
        data: dict[str, t.Any],
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Update a given group.

        :param group_id: the ID of the group
        :param data: the group document to use for update
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PUT /v2/groups/<group_id>``

                .. extdoclink:: Update a Group
                    :service: groups
                    :ref: update_group_v2_groups__group_id__put
        """
        return self.put(f"/v2/groups/{group_id}", data=data, query_params=query_params)

    def get_group_policies(
        self,
        group_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get policies for the given group.

        :param group_id: the ID of the group
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /v2/groups/<group_id>/policies``

                .. extdoclink:: Get the policies for a group
                    :service: groups
                    :ref: get_policies_v2_groups__group_id__policies_get
        """
        return self.get(f"/v2/groups/{group_id}/policies", query_params=query_params)

    def set_group_policies(
        self,
        group_id: uuid.UUID | str,
        data: dict[str, t.Any] | GroupPolicies,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Set policies for the group.

        :param group_id: the ID of the group
        :param data: the group policy document to set
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PUT /v2/groups/<group_id>/policies``

                .. extdoclink:: Set the policies for a group
                    :service: groups
                    :ref: update_policies_v2_groups__group_id__policies_put
        """
        return self.put(
            f"/v2/groups/{group_id}/policies", data=data, query_params=query_params
        )

    def get_identity_preferences(
        self, *, query_params: dict[str, t.Any] | None = None
    ) -> response.GlobusHTTPResponse:
        """
        Get identity preferences.  Currently this only includes whether the
        user allows themselves to be added to groups.

        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /v2/preferences``

                .. extdoclink:: Get the preferences for your identity set
                    :service: groups
                    :ref: get_identity_set_preferences_v2_preferences_get
        """
        return self.get("/v2/preferences", query_params=query_params)

    def set_identity_preferences(
        self,
        data: dict[str, t.Any],
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Set identity preferences.  Currently this only includes whether the
        user allows themselves to be added to groups.

        :param data: the identity set preferences document
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    gc = globus_sdk.GroupsClient(...)
                    gc.set_identity_preferences({"allow_add": False})

            .. tab-item:: API Info

                ``PUT /v2/preferences``

                .. extdoclink:: Set the preferences for your identity set
                    :service: groups
                    :ref: put_identity_set_preferences_v2_preferences_put
        """
        return self.put("/v2/preferences", data=data, query_params=query_params)

    def get_membership_fields(
        self,
        group_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get membership fields for your identities.

        :param group_id: the ID of the group
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /v2/groups/<group_id>/membership_fields``

                .. extdoclink:: Get the membership fields for your identity set
                    :service: groups
                    :ref: get_membership_fields_v2_groups__group_id__membership_fields_get
        """  # noqa: E501
        return self.get(
            f"/v2/groups/{group_id}/membership_fields", query_params=query_params
        )

    def set_membership_fields(
        self,
        group_id: uuid.UUID | str,
        data: dict[t.Any, str],
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Set membership fields for your identities.

        :param group_id: the ID of the group
        :param data: the membership fields document
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PUT /v2/groups/<group_id>/membership_fields``

                .. extdoclink:: Set the membership fields for your identity set
                    :service: groups
                    :ref: put_membership_fields_v2_groups__group_id__membership_fields_put
        """  # noqa: E501
        return self.put(
            f"/v2/groups/{group_id}/membership_fields",
            data=data,
            query_params=query_params,
        )

    def batch_membership_action(
        self,
        group_id: uuid.UUID | str,
        actions: dict[str, t.Any] | BatchMembershipActions,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Execute a batch of actions against several group memberships.

        :param group_id: the ID of the group
        :param actions: the batch of membership actions to perform, modifying, creating,
            and removing memberships in the group
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    gc = globus_sdk.GroupsClient(...)
                    group_id = ...
                    batch = globus_sdk.BatchMembershipActions()
                    batch.add_members("ae332d86-d274-11e5-b885-b31714a110e9")
                    batch.invite_members("c699d42e-d274-11e5-bf75-1fc5bf53bb24")
                    gc.batch_membership_action(group_id, batch)

            .. tab-item:: API Info

                ``PUT /v2/groups/<group_id>/membership_fields``

                .. extdoclink:: Perform actions on members of the group
                    :service: groups
                    :ref: group_membership_post_actions_v2_groups__group_id__post
        """
        return self.post(
            f"/v2/groups/{group_id}", data=actions, query_params=query_params
        )

    def set_subscription_admin_verified(
        self,
        group_id: uuid.UUID | str,
        subscription_id: uuid.UUID | str | None,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Verify a group as belonging to a subscription or disassociate a verified group
            from a subscription.

        :param group_id: the ID of the group
        :param subscription_id: the ID of the subscription to which the group belongs,
            or ``None`` to disassociate the group from a subscription
        :param query_params: additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PUT /v2/groups/<group_id>/subscription_admin_verified``

                .. extdoclink:: Update the group's subscription admin verified ID
                    :service: groups
                    :ref: update_subscription_admin_verified_id_v2_groups__group_id__
                        subscription_admin_verified_put
        """
        return self.put(
            f"/v2/groups/{group_id}/subscription_admin_verified",
            data={"subscription_admin_verified_id": subscription_id},
            query_params=query_params,
        )



from __future__ import annotations

from globus_sdk._internal import guards
from globus_sdk.exc import ErrorSubdocument, GlobusAPIError


class FlowsAPIError(GlobusAPIError):
    """
    Error class to represent error responses from Flows.
    """

    def _parse_undefined_error_format(self) -> bool:
        """
        Treat any top-level "error" key as an "array of size 1".
        Meaning that we'll see a single subdocument for data shaped like
            {
              "error": {
                "foo": "bar"
              }
            }
        """
        # if there is not a top-level 'error' key, no special behavior is defined
        # fall-back to the base class implementation
        if not isinstance(self._dict_data.get("error"), dict):
            return super()._parse_undefined_error_format()

        self.errors = [ErrorSubdocument(self._dict_data["error"])]
        self.code = self._extract_code_from_error_array(self.errors)

        details = self._dict_data["error"].get("detail")
        if guards.is_list_of(details, dict):
            self.messages = [
                error_detail["msg"]
                for error_detail in details
                if isinstance(error_detail.get("msg"), str)
            ]
        else:
            self.messages = self._extract_messages_from_error_array(self.errors)

        return True



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import TASK_DOC, TASK_ID

RESPONSES = ResponseSet(
    metadata={"task_id": TASK_ID},
    default=RegisteredResponse(
        service="compute",
        path=f"/v2/tasks/{TASK_ID}",
        method="GET",
        json=TASK_DOC,
    ),
)



from __future__ import annotations

import argparse
import time
import typing as t

import boto3

import globus_sdk
from globus_sdk.token_storage import StorageAdapter

CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"
tablename = "example-globus-tokenstorage"

parser = argparse.ArgumentParser()
parser.add_argument(
    "--create",
    action="store_true",
)

boto_client = boto3.client("dynamodb")
auth_client = globus_sdk.NativeAppAuthClient(CLIENT_ID)


class DynamoDBStorageAdapter(StorageAdapter):
    def __init__(self, client, tablename: str, namespace: str = "DEFAULT") -> None:
        """
        :param client: A boto3 DyanmoDB client to use
        :param tablename: The name of the dynamodb table to use
        :param namespace: A namespace for all keys within this storage.
            Setting up explicit namespacing allows for multiple storage
            adapters for multiple users or applications to share a table.
        """
        self.client = client
        self.tablename = tablename
        self.namespace = namespace

    def _compute_key(self, resource_server: str) -> str:
        """
        Compute the 'token_data_id' used for storage and retrieval
        vis-a-vis a specific resource_server/namespace combination.

        This is defined as a simple delimited string which starts with the
        namespace given.

        Globus keys tokens by the ``resource_server`` string, but also has
        additional context about which user and application were being
        used. For the storage adapter, we will need to use namespacing to
        separate users.

        Consider setting ``namespace`` to a value like a user ID or a
        combination of user ID and authentication context.
        """
        return f"{self.namespace}:{resource_server}"

    def store(self, token_response: globus_sdk.OAuthTokenResponse) -> None:
        for resource_server, token_data in token_response.by_resource_server.items():
            key = self._compute_key(resource_server)
            dynamo_item = {
                "token_data_id": {"S": key},
                "resource_server": {"S": resource_server},
                "access_token": {"S": token_data["access_token"]},
                "refresh_token": {"S": token_data["refresh_token"]},
                "expires_at_seconds": {"N": str(token_data["expires_at_seconds"])},
                "scope": {"S": token_data["scope"]},
            }
            # avoid setting `refresh_token` if it is null (meaning the
            # login flow used access tokens only)
            if token_data["refresh_token"] is None:
                del dynamo_item["refresh_token"]

            self.client.put_item(TableName=self.tablename, Item=dynamo_item)

    def get_token_data(self, resource_server: str) -> dict[str, t.Any] | None:
        key = self._compute_key(resource_server)

        wrapped_item = self.client.get_item(
            TableName=self.tablename,
            Key={"token_data_id": {"S": key}},
            ConsistentRead=True,
        )
        if "Item" not in wrapped_item:
            return None

        dynamo_item = wrapped_item["Item"]
        return {
            "resource_server": dynamo_item["resource_server"]["S"],
            "access_token": dynamo_item["access_token"]["S"],
            "refresh_token": dynamo_item.get("refresh_token", {"S": None})["S"],
            "expires_at_seconds": int(dynamo_item["expires_at_seconds"]["N"]),
            "scope": dynamo_item["scope"]["S"],
        }


def create_table():
    # create a table with a key of "token_data_id"
    # this is a nonspecific string key which we will compute
    #
    # the relationship of "token_data_id" to the token will be explained below
    boto_client.create_table(
        TableName=tablename,
        KeySchema=[{"AttributeName": "token_data_id", "KeyType": "HASH"}],
        AttributeDefinitions=[{"AttributeName": "token_data_id", "AttributeType": "S"}],
        BillingMode="PROVISIONED",
        ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
    )
    print(f"'{tablename}' create in progress.")

    # poll status until the table is "ACTIVE"
    print("Waiting for active status (Ctrl+C to cancel)...", end="", flush=True)
    status = None
    while status != "ACTIVE":
        time.sleep(1)
        try:
            r = boto_client.describe_table(TableName=tablename)
        except boto_client.exceptions.ResourceNotFoundException:
            continue
        print(".", end="", flush=True)
        status = r["Table"]["TableStatus"]
    print("ok")


def do_login_flow(storage: StorageAdapter):
    auth_client.oauth2_start_flow(
        requested_scopes=globus_sdk.GroupsClient.scopes.view_my_groups_and_memberships,
        refresh_tokens=True,
    )
    authorize_url = auth_client.oauth2_get_authorize_url()
    print(f"Please go to this URL and login:\n\n{authorize_url}\n")
    auth_code = input("Please enter the code here: ").strip()
    tokens = auth_client.oauth2_exchange_code_for_tokens(auth_code)
    storage.store(tokens)
    return tokens.by_resource_server[globus_sdk.GroupsClient.resource_server]


def group_list(storage: StorageAdapter):
    tokens = storage.get_token_data(globus_sdk.GroupsClient.resource_server)
    if tokens is None:
        tokens = do_login_flow(storage)

    authorizer = globus_sdk.RefreshTokenAuthorizer(
        tokens["refresh_token"],
        auth_client,
        access_token=tokens["access_token"],
        expires_at=tokens["expires_at_seconds"],
        on_refresh=storage.store,
    )
    groups_client = globus_sdk.GroupsClient(authorizer=authorizer)
    print("ID,Name,Type,Session Enforcement,Roles")
    for group in groups_client.get_my_groups():
        # parse the group to get data for output
        if group.get("enforce_session"):
            session_enforcement = "strict"
        else:
            session_enforcement = "not strict"
        roles = ",".join({m["role"] for m in group["my_memberships"]})

        print(
            ",".join(
                [
                    group["id"],
                    group["name"],
                    group["group_type"],
                    session_enforcement,
                    roles,
                ]
            )
        )


if __name__ == "__main__":
    args = parser.parse_args()
    if args.create:
        create_table()
    else:
        storage = DynamoDBStorageAdapter(boto_client, tablename)
        group_list(storage)



import pytest

from globus_sdk import AuthAPIError
from globus_sdk.testing import construct_error


def test_auth_error_get_args_simple():
    err = construct_error(
        error_class=AuthAPIError,
        http_status=404,
        body={"detail": "simple auth error message"},
    )
    req = err._underlying_response.request

    assert err._get_args() == [
        req.method,
        req.url,
        None,
        404,
        None,
        "simple auth error message",
    ]


def test_nested_auth_error_message_and_code():
    err = construct_error(
        error_class=AuthAPIError,
        http_status=404,
        body={
            "errors": [
                {"detail": "nested auth error message", "code": "Auth Error"},
                {
                    "title": "some secondary error",
                    "code": "HiddenError",
                },
            ]
        },
    )

    assert err.message == "nested auth error message; some secondary error"
    assert err.code is None


@pytest.mark.parametrize(
    "error_body, expected_error_id",
    (
        (
            {
                "errors": [
                    {"id": "foo"},
                ]
            },
            "foo",
        ),
        (
            {
                "errors": [
                    {"id": "foo"},
                    {"id": "bar"},
                ]
            },
            None,
        ),
        (
            {
                "errors": [
                    {"id": "foo"},
                    {"id": "foo"},
                ]
            },
            "foo",
        ),
        (
            {"errors": [{"id": "foo"}, {"id": "foo"}, {}]},
            "foo",
        ),
    ),
)
def test_auth_error_parses_error_id(error_body, expected_error_id):
    err = construct_error(
        error_class=AuthAPIError,
        http_status=404,
        body=error_body,
    )
    assert err.request_id == expected_error_id



from unittest import mock

from globus_sdk.transport import (
    RequestCallerInfo,
    RetryCheckResult,
    RetryCheckRunner,
    RetryConfig,
    RetryContext,
)
from globus_sdk.transport.default_retry_checks import DEFAULT_RETRY_CHECKS


def _make_test_retry_context(*, status=200, exception=None, response=None):
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(DEFAULT_RETRY_CHECKS)
    caller_info = RequestCallerInfo(retry_config=retry_config)
    if exception:
        return RetryContext(1, caller_info=caller_info, exception=exception)
    elif response:
        return RetryContext(1, caller_info=caller_info, response=response)

    dummy_response = mock.Mock()
    dummy_response.status_code = 200
    return RetryContext(1, caller_info=caller_info, response=dummy_response)


def test_retry_check_runner_should_retry_explicit_on_first_check():
    def check1(ctx):
        return RetryCheckResult.do_not_retry

    def check2(ctx):
        return RetryCheckResult.do_retry

    failing_checker = RetryCheckRunner([check1, check2])
    assert failing_checker.should_retry(_make_test_retry_context()) is False
    passing_checker = RetryCheckRunner([check2, check1])
    assert passing_checker.should_retry(_make_test_retry_context()) is True


def test_retry_check_runner_fallthrough_to_false():
    def check1(ctx):
        return RetryCheckResult.no_decision

    def check2(ctx):
        return RetryCheckResult.no_decision

    checker = RetryCheckRunner([check1, check2])
    assert checker.should_retry(_make_test_retry_context()) is False



from __future__ import annotations

import pytest

from globus_sdk import GlobusSDKUsageError
from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "case_name",
    (
        "name",
        "public_client",
        "private_client",
        "publicly_visible",
        "not_publicly_visible",
        "redirect_uris",
        "links",
        "required_idp",
        "preselect_idp",
        "client_type_confidential_client",
        "client_type_public_installed_client",
        "client_type_client_identity",
        "client_type_resource_server",
        "client_type_globus_connect_server",
        "client_type_hybrid_confidential_client_resource_server",
        "client_type_public_webapp_client",
    ),
)
def test_create_child_client_args(
    auth_client,
    case_name: str,
):
    meta = load_response(auth_client.create_child_client, case=case_name).metadata

    res = auth_client.create_child_client(**meta["args"])
    for k, v in meta["response"].items():
        assert res["client"][k] == v


def test_links_requirement(auth_client):
    """
    Verify that terms_and_conditions and privacy_policy must be used together.
    """
    with pytest.raises(GlobusSDKUsageError):
        auth_client.create_child_client(
            "FOO",
            visibility="public",
            terms_and_conditions="https://foo.net",
        )

    with pytest.raises(GlobusSDKUsageError):
        auth_client.create_child_client(
            "FOO",
            visibility="public",
            privacy_policy="https://foo.net",
        )


def test_public_client_and_client_type_requirement(auth_client):
    """
    Verify that exactly one of ``public_client` and ``client_type`` are expected.
    """
    # Neither public_client nor client_type
    with pytest.raises(GlobusSDKUsageError):
        auth_client.create_child_client(
            "FOO",
        )

    # Both public_client and client_type
    with pytest.raises(GlobusSDKUsageError):
        auth_client.create_child_client(
            "FOO",
            public_client=True,
            client_type="GCS",
        )



import uuid
from unittest import mock

import pytest
import responses

import globus_sdk
from globus_sdk.experimental.gcs_downloader import GCSDownloader


def test_collection_id_sniffing():
    file_url = "https://some-base-url.example.com/bogus.txt"
    client_id = str(uuid.UUID(int=3))

    responses.add(
        method="GET",
        url=file_url,
        status=302,
        headers={
            "Location": f"https://auth.globus.org/authenticate?client_id={client_id}"
        },
    )

    downloader = GCSDownloader(app=mock.Mock())
    assert downloader._sniff_collection_id(file_url) == client_id


@pytest.mark.parametrize(
    "entity_type", ("GCSv5_mapped_collection", "GCSv5_guest_collection")
)
@pytest.mark.parametrize("high_assurance", (True, False))
def test_required_scope_detection(entity_type, high_assurance):
    gcs_client = mock.Mock()
    gcs_client.collection_id = str(uuid.UUID(int=8))
    scopes = gcs_client.scopes = globus_sdk.scopes.GCSCollectionScopes(
        gcs_client.collection_id
    )

    transfer_client = mock.Mock()
    transfer_client.get_endpoint.return_value = {
        "entity_type": entity_type,
        "high_assurance": high_assurance,
    }

    downloader = GCSDownloader(mock.Mock(), transfer_client=transfer_client)

    result = downloader._determine_required_scopes(gcs_client)
    if "mapped" in entity_type and not high_assurance:
        assert {str(s) for s in result} == {str(scopes.https), str(scopes.data_access)}
    else:
        assert {str(s) for s in result} == {str(scopes.https)}



import pytest

import globus_sdk
from globus_sdk.testing import load_response


def test_update_index(client):
    meta = load_response(client.update_index).metadata

    res = client.update_index(meta["index_id"], display_name="foo")
    assert res.http_status == 200
    assert res["display_name"] == meta["display_name"]


def test_update_index_forbidden_error(client):
    meta = load_response(client.update_index, case="forbidden").metadata

    with pytest.raises(globus_sdk.SearchAPIError) as excinfo:
        client.update_index(meta["index_id"])

    err = excinfo.value

    assert err.http_status == 403
    assert err.code == "Forbidden.Generic"



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        status=200,
        json={
            "access_token": "transfer_access_token",
            "scope": "urn:globus:auth:scope:transfer.api.globus.org:all",
            "expires_in": 172800,
            "token_type": "Bearer",
            "resource_server": "transfer.api.globus.org",
            "state": "_default",
            "other_tokens": [],
        },
    ),
    invalid_grant=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        status=401,
        json={"error": "invalid_grant"},
    ),
    openid=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        status=200,
        json={
            "access_token": "auth_access_token",
            "scope": "openid",
            "expires_in": 172800,
            "token_type": "Bearer",
            "resource_server": "auth.globus.org",
            "id_token": "openid_token",
            "other_tokens": [],
        },
    ),
)



import pytest

import globus_sdk
from globus_sdk.testing import load_response


def test_delete_index(client):
    meta = load_response(client.delete_index).metadata

    res = client.delete_index(meta["index_id"])
    assert res.http_status == 200
    assert res["acknowledged"] is True


def test_delete_index_delete_already_pending(client):
    meta = load_response(client.delete_index, case="delete_pending").metadata

    with pytest.raises(globus_sdk.SearchAPIError) as excinfo:
        client.delete_index(meta["index_id"])

    err = excinfo.value

    assert err.http_status == 409
    assert err.code == "Conflict.IncompatibleIndexStatus"



from unittest import mock

import pytest

from globus_sdk.authorizers import NullAuthorizer
from globus_sdk.transport import RequestCallerInfo, RequestsTransport, RetryConfig


def test_will_not_modify_authz_header_without_authorizer():
    request = mock.Mock()
    request.headers = {}

    transport = RequestsTransport()
    transport._set_authz_header(None, request)
    assert request.headers == {}

    request.headers["Authorization"] = "foo bar"
    transport._set_authz_header(None, request)
    assert request.headers == {"Authorization": "foo bar"}


def test_will_null_authz_header_with_null_authorizer():
    request = mock.Mock()
    request.headers = {}

    transport = RequestsTransport()
    transport._set_authz_header(NullAuthorizer(), request)
    assert request.headers == {}

    request.headers["Authorization"] = "foo bar"
    transport._set_authz_header(NullAuthorizer(), request)
    assert request.headers == {}


def test_requests_transport_accepts_caller_info():
    retry_config = RetryConfig()
    transport = RequestsTransport()
    mock_authorizer = mock.Mock()
    mock_authorizer.get_authorization_header.return_value = "Bearer token"
    caller_info = RequestCallerInfo(
        retry_config=retry_config, authorizer=mock_authorizer
    )

    with mock.patch.object(transport, "session") as mock_session:
        mock_response = mock.Mock(status_code=200)
        mock_session.send.return_value = mock_response

        response = transport.request(
            "GET", "https://example.com", caller_info=caller_info
        )

        assert response.status_code == 200

        sent_request = mock_session.send.call_args[0][0]
        assert sent_request.headers["Authorization"] == "Bearer token"


def test_requests_transport_caller_info_required():
    transport = RequestsTransport()

    with pytest.raises(TypeError):
        transport.request("GET", "https://example.com")


def test_requests_transport_keyword_only():
    retry_config = RetryConfig()
    transport = RequestsTransport()
    caller_info = RequestCallerInfo(retry_config=retry_config)

    with pytest.raises(TypeError):
        transport.request("GET", "https://example.com", caller_info)



import inspect
import os

import responses

from globus_sdk._internal.utils import slash_join


def register_api_route_fixture_file(service, path, filename, **kwargs):
    """
    register an API route to serve the contents of a file, given the name of
    that file in a `fixture_data` directory, adjacent to the current (calling)
    module

    i.e. in a dir like this:
      path/to/tests
      ├── test_mod.py
      └── fixture_data
          └── dat.txt

    you can call
    >>> register_api_route_fixture_file('transfer', '/foo', 'dat.txt')

    in `test_mod.py`

    it will "do the right thing" and find the abspath to dat.txt , and
    load the contents of that file as the response for '/foo'
    """
    # get calling frame
    frm = inspect.stack()[1]
    # get filename from frame, make it absolute
    modpath = os.path.abspath(frm[1])

    abspath = os.path.join(os.path.dirname(modpath), "fixture_data", filename)
    with open(abspath, "rb") as f:
        body = f.read()

    register_api_route(service, path, body=body, **kwargs)


def register_api_route(
    service, path, method=responses.GET, adding_headers=None, replace=False, **kwargs
):
    """
    Handy wrapper for adding URIs to the response mock state.
    """
    base_url_map = {
        "auth": "https://auth.globus.org/",
        "nexus": "https://nexus.api.globusonline.org/",
        "groups": "https://groups.api.globus.org/",
        "transfer": "https://transfer.api.globus.org/v0.10",
        "search": "https://search.api.globus.org/",
        "gcs": "https://abc.xyz.data.globus.org/api/",
    }
    assert service in base_url_map
    base_url = base_url_map.get(service)
    full_url = slash_join(base_url, path)

    # can set it to `{}` explicitly to clear the default
    if adding_headers is None:
        adding_headers = {"Content-Type": "application/json"}

    if replace:
        responses.replace(
            method, full_url, headers=adding_headers, match_querystring=None, **kwargs
        )
    else:
        responses.add(
            method, full_url, headers=adding_headers, match_querystring=None, **kwargs
        )



import copy

from globus_sdk.testing import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_RUN

RUN_ID = TWO_HOP_TRANSFER_RUN["run_id"]
SUCCESS_RESPONSE = copy.deepcopy(TWO_HOP_TRANSFER_RUN)
SUCCESS_RESPONSE.update(
    {
        "status": "FAILED",
        "display_status": "FAILED",
        "details": {
            "time": "2023-06-12T23:04:42.121000+00:00",
            "code": "FlowCanceled",
            "description": "The Flow Instance was canceled by the user",
            "details": {},
        },
    }
)

RESPONSES = ResponseSet(
    metadata={"run_id": RUN_ID},
    default=RegisteredResponse(
        service="flows",
        method="POST",
        path=f"/runs/{RUN_ID}/cancel",
        json=SUCCESS_RESPONSE,
    ),
)



from unittest import mock

from globus_sdk.services.transfer.transport import TRANSFER_DEFAULT_RETRY_CHECKS
from globus_sdk.transport import (
    RequestCallerInfo,
    RetryCheckCollection,
    RetryCheckRunner,
    RetryConfig,
    RetryContext,
)
from globus_sdk.transport.default_retry_checks import DEFAULT_RETRY_CHECKS


def test_transfer_only_replaces_checks():
    # their length matches, meaning things line up
    assert len(TRANSFER_DEFAULT_RETRY_CHECKS) == len(DEFAULT_RETRY_CHECKS)

    # also confirm that this holds once loaded
    # if the implementation of the RetryCheckCollection becomes sensitive to
    # the contents of these tuples, this could fail
    default_variant = RetryCheckCollection()
    default_variant.register_many_checks(DEFAULT_RETRY_CHECKS)

    transfer_variant = RetryCheckCollection()
    transfer_variant.register_many_checks(TRANSFER_DEFAULT_RETRY_CHECKS)

    assert len(default_variant) == len(transfer_variant)


def test_transfer_does_not_retry_external():
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(TRANSFER_DEFAULT_RETRY_CHECKS)
    checker = RetryCheckRunner(retry_config.checks)

    body = {
        "HTTP status": "502",
        "code": "ExternalError.DirListingFailed.GCDisconnected",
        "error_name": "Transfer API Error",
        "message": "The GCP endpoint is not currently connected to Globus",
        "request_id": "rhvcR0aHX",
    }

    dummy_response = mock.Mock()
    dummy_response.json = lambda: body
    dummy_response.status_code = 502
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, response=dummy_response)

    assert checker.should_retry(ctx) is False


def test_transfer_does_not_retry_endpoint_error():
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(TRANSFER_DEFAULT_RETRY_CHECKS)
    checker = RetryCheckRunner(retry_config.checks)

    body = {
        "HTTP status": "502",
        "code": "EndpointError",
        "error_name": "Transfer API Error",
        "message": (
            "This GCSv5 is older than version 5.4.62 and does not support local user "
            "selection"
        ),
        "request_id": "istNh0Zpz",
    }

    dummy_response = mock.Mock()
    dummy_response.json = lambda: body
    dummy_response.status_code = 502
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, response=dummy_response)

    assert checker.should_retry(ctx) is False


def test_transfer_retries_others():
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(TRANSFER_DEFAULT_RETRY_CHECKS)
    checker = RetryCheckRunner(retry_config.checks)

    def _raise_value_error():
        raise ValueError()

    dummy_response = mock.Mock()
    dummy_response.json = _raise_value_error
    dummy_response.status_code = 502
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, response=dummy_response)

    assert checker.should_retry(ctx) is True



"""
Various tests for imports from the SDK to ensure that our `__all__`
declarations are correct and complete.

To ensure that we aren't contaminating these tests with the current
interpreter's state, invoke these via subprocess check_call() calls.
These should all be using `shell=True` (the subprocess invoked interpreter
doesn't behave correctly without it).
"""

import os
import subprocess
import sys

import pytest

PYTHON_BINARY = os.environ.get("GLOBUS_TEST_PY", sys.executable)


@pytest.mark.parametrize(
    "importstring",
    [
        "from globus_sdk import *",
        "from globus_sdk import TransferClient, AuthClient, SearchClient",
    ],
)
def test_import_str(importstring):
    proc = subprocess.Popen(
        f'{PYTHON_BINARY} -c "{importstring}"',
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    status = proc.wait()
    assert status == 0, str(proc.communicate())
    proc.stdout.close()
    proc.stderr.close()



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import ENDPOINT_ID

RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="transfer",
        method="PUT",
        path=f"/v0.10/endpoint/{ENDPOINT_ID}",
        json={
            "DATA_TYPE": "result",
            "code": "Updated",
            "message": "Endpoint updated successfully",
            "request_id": "6aZjzldyM",
            "resource": f"/v0.10/endpoint/{ENDPOINT_ID}",
        },
    ),
)



from __future__ import annotations

import logging
import sys
import typing as t
import uuid

from globus_sdk import GlobusHTTPResponse, client, paging
from globus_sdk._internal import guards
from globus_sdk._internal.remarshal import commajoin
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.authorizers import GlobusAuthorizer
from globus_sdk.globus_app import GlobusApp
from globus_sdk.scopes import (
    FlowsScopes,
    GCSCollectionScopes,
    Scope,
    SpecificFlowScopes,
    TransferScopes,
)
from globus_sdk.transport import RequestsTransport, RetryConfig

from .data import RunActivityNotificationPolicy
from .errors import FlowsAPIError
from .response import (
    IterableFlowsResponse,
    IterableRegisteredAPIsResponse,
    IterableRunLogsResponse,
    IterableRunsResponse,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

log = logging.getLogger(__name__)


class FlowsClient(client.BaseClient):
    r"""
    Client for the Globus Flows API.

    .. sdk-sphinx-copy-params:: BaseClient

    .. automethodlist:: globus_sdk.FlowsClient
    """

    error_class = FlowsAPIError
    service_name = "flows"
    scopes = FlowsScopes
    default_scope_requirements = [FlowsScopes.all]

    def create_flow(
        self,
        title: str,
        definition: dict[str, t.Any],
        input_schema: dict[str, t.Any],
        subtitle: str | MissingType = MISSING,
        description: str | MissingType = MISSING,
        flow_viewers: list[str] | MissingType = MISSING,
        flow_starters: list[str] | MissingType = MISSING,
        flow_administrators: list[str] | MissingType = MISSING,
        run_managers: list[str] | MissingType = MISSING,
        run_monitors: list[str] | MissingType = MISSING,
        keywords: list[str] | MissingType = MISSING,
        subscription_id: uuid.UUID | str | None | MissingType = MISSING,
        authentication_policy_id: uuid.UUID | str | None | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        Create a flow

        :param title: A non-unique, human-friendly name used for displaying the
            flow to end users. (1 - 128 characters)
        :param definition: JSON object specifying flows states and execution order. For
            a more detailed explanation of the flow definition, see
            `Authoring Flows <https://docs.globus.org/api/flows/authoring-flows>`_
        :param input_schema: A JSON Schema to which flow run input must conform
        :param subtitle: A concise summary of the flow’s purpose. (0 - 128 characters)
        :param description: A detailed description of the flow's purpose for end user
            display. (0 - 4096 characters)
        :param flow_viewers: A set of Principal URN values, or the value "public",
            indicating entities who can view the flow

            .. dropdown:: Example Values

                .. code-block:: json

                    [ "public" ]

                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]


        :param flow_starters: A set of Principal URN values, or the value
            "all_authenticated_users", indicating entities who can initiate a *run* of
            the flow

            .. dropdown:: Example Values

                .. code-block:: json

                    [ "all_authenticated_users" ]


                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param flow_administrators: A set of Principal URN values indicating entities
            who can perform administrative operations on the flow (create, delete,
            update)

            .. dropdown:: Example Values

                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param run_managers: A set of Principal URN values indicating entities who can
            perform management operations on the flow's *runs*.

            .. dropdown:: Example Values

                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param run_monitors: A set of Principal URN values indicating entities who can
            monitor the flow's *runs*.

            .. dropdown:: Example Values

                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param keywords: A set of terms used to categorize the flow used in query and
            discovery operations (0 - 1024 items)
        :param subscription_id: The ID of the subscription to associate with the flow,
            marking as a subscription tier flow.
        :param authentication_policy_id: The ID of the authentication policy to associate
             with the flow.
        :param additional_fields: Additional Key/Value pairs sent to the create API

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    ...
                    flows = FlowsClient(...)
                    flows.create_flow(
                        title="my-cool-flow",
                        definition={
                            "StartAt": "the-one-true-state",
                            "States": {"the-one-true-state": {"Type": "Pass", "End": True}},
                        },
                        input_schema={
                            "type": "object",
                            "properties": {
                                "input-a": {"type": "string"},
                                "input-b": {"type": "number"},
                                "input-c": {"type": "boolean"},
                            },
                        },
                    )

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.create_flow

            .. tab-item:: API Info

                .. extdoclink:: Create Flow
                    :service: flows
                    :ref: Flows/paths/~1flows/post
        """  # noqa E501

        data = {
            "title": title,
            "definition": definition,
            "input_schema": input_schema,
            "subtitle": subtitle,
            "description": description,
            "flow_viewers": flow_viewers,
            "flow_starters": flow_starters,
            "flow_administrators": flow_administrators,
            "run_managers": run_managers,
            "run_monitors": run_monitors,
            "keywords": keywords,
            "subscription_id": subscription_id,
            "authentication_policy_id": authentication_policy_id,
            **(additional_fields or {}),
        }
        return self.post("/flows", data=data)

    def get_flow(
        self,
        flow_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """Retrieve a flow by ID

        :param flow_id: The ID of the flow to fetch
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Flow
                    :service: flows
                    :ref: Flows/paths/~1flows~1{flow_id}/get
        """
        return self.get(f"/flows/{flow_id}", query_params=query_params)

    @paging.has_paginator(paging.MarkerPaginator, items_key="flows")
    def list_flows(
        self,
        *,
        filter_roles: str | t.Iterable[str] | MissingType = MISSING,
        filter_fulltext: str | MissingType = MISSING,
        orderby: str | t.Iterable[str] | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableFlowsResponse:
        """
        List deployed flows

        :param filter_roles: A list of role names specifying the roles the user must
            have for a flow to be included in the response.
        :param filter_fulltext: A string to use in a full-text search to filter results
        :param orderby: A criterion for ordering flows in the listing
        :param marker: A marker for pagination
        :param query_params: Any additional parameters to be passed through
            as query params.

        **Role Filters**

        ``filter_roles`` accepts a list of roles which are used to filter the results to
        flows where the caller has any of the specified roles.

        The valid role values are:

        - ``flow_viewer``
        - ``flow_starter``
        - ``flow_administrator``
        - ``flow_owner``
        - ``run_monitor``
        - ``run_manager``

        **OrderBy Values**

        Values for ``orderby`` consist of a field name, a space, and an
        ordering mode -- ``ASC`` for "ascending" and ``DESC`` for "descending".

        Supported field names are

          - ``id``
          - ``scope_string``
          - ``flow_owners``
          - ``flow_administrators``
          - ``title``
          - ``created_at``
          - ``updated_at``

        For example, ``orderby="updated_at DESC"`` requests a descending sort on update
        times, getting the most recently updated flow first. Multiple ``orderby`` values
        may be given as an iterable, e.g. ``orderby=["updated_at DESC", "title ASC"]``.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    import json
                    import textwrap

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    my_frobulate_flows = flows.list_flows(
                        filter_roles="flow_owner",
                        filter_fulltext="frobulate",
                        orderby=("title ASC", "updated_at DESC"),
                    )
                    for flow_doc in my_frobulate_flows:
                        print(f"Title: {flow_doc['title']}")
                        print(f"Description: {flow_doc['description']}")
                        print("Definition:")
                        print(
                            textwrap.indent(
                                json.dumps(
                                    flow_doc["definition"],
                                    indent=2,
                                    separators=(",", ": "),
                                ),
                                "    ",
                            )
                        )
                        print()

            .. tab-item:: Paginated Usage

                .. paginatedusage:: list_flows

            .. tab-item:: API Info

                .. extdoclink:: List Flows
                    :service: flows
                    :ref: Flows/paths/~1flows/get
        """
        query_params = {
            "filter_roles": commajoin(filter_roles),
            "filter_fulltext": filter_fulltext,
            # if `orderby` is an iterable (e.g., generator expression), it gets
            # converted to a list in this step
            "orderby": (
                orderby if isinstance(orderby, (str, MissingType)) else list(orderby)
            ),
            "marker": marker,
            **(query_params or {}),
        }
        return IterableFlowsResponse(self.get("/flows", query_params=query_params))

    def update_flow(
        self,
        flow_id: uuid.UUID | str,
        *,
        title: str | MissingType = MISSING,
        definition: dict[str, t.Any] | MissingType = MISSING,
        input_schema: dict[str, t.Any] | MissingType = MISSING,
        subtitle: str | MissingType = MISSING,
        description: str | MissingType = MISSING,
        flow_owner: str | MissingType = MISSING,
        flow_viewers: list[str] | MissingType = MISSING,
        flow_starters: list[str] | MissingType = MISSING,
        flow_administrators: list[str] | MissingType = MISSING,
        run_managers: list[str] | MissingType = MISSING,
        run_monitors: list[str] | MissingType = MISSING,
        keywords: list[str] | MissingType = MISSING,
        subscription_id: uuid.UUID | str | t.Literal["DEFAULT"] | MissingType = MISSING,
        authentication_policy_id: uuid.UUID | str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        Update a flow

        Only the parameter `flow_id` is required.
        Any fields omitted from the request will be unchanged

        :param flow_id: The ID of the flow to fetch
        :param title: A non-unique, human-friendly name used for displaying the
            flow to end users. (1 - 128 characters)
        :param definition: JSON object specifying flows states and execution order. For
            a more detailed explanation of the flow definition, see
            `Authoring Flows <https://docs.globus.org/api/flows/authoring-flows>`_
        :param input_schema: A JSON Schema to which flow run input must conform
        :param subtitle: A concise summary of the flow’s purpose. (0 - 128 characters)
        :param description: A detailed description of the flow's purpose for end user
            display. (0 - 4096 characters)
        :param flow_owner: An Auth Identity URN to set as flow owner; this must match
            the Identity URN of the entity calling `update_flow`
        :param flow_viewers: A set of Principal URN values, or the value "public",
            indicating entities who can view the flow

            .. dropdown:: Example Values

                .. code-block:: json

                    [ "public" ]

                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param flow_starters: A set of Principal URN values, or the value
            "all_authenticated_users", indicating entities who can initiate a *run* of
            the flow

            .. dropdown:: Example Values

                .. code-block:: json

                    [ "all_authenticated_users" ]


                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param flow_administrators: A set of Principal URN values indicating entities
            who can perform administrative operations on the flow (create, delete,
            update)

            .. dropdown:: Example Value

                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param run_managers: A set of Principal URN values indicating entities who can
            perform management operations on the flow's *runs*.

            .. dropdown:: Example Values

                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param run_monitors: A set of Principal URN values indicating entities who can
            monitor the flow's *runs*.

            .. dropdown:: Example Values

                .. code-block:: json

                    [
                        "urn:globus:auth:identity:b44bddda-d274-11e5-978a-9f15789a8150",
                        "urn:globus:groups:id:c1dcd951-3f35-4ea3-9f28-a7cdeaf8b68f"
                    ]

        :param keywords: A set of terms used to categorize the flow used in query and
            discovery operations (0 - 1024 items)
        :param subscription_id: A subscription ID to assign to the flow.
        :param authentication_policy_id: An authentication policy to assign to the flow.
        :param additional_fields: Additional Key/Value pairs sent to the create API

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    flows.update_flow(
                        flow_id="581753c7-45da-43d3-ad73-246b46e7cb6b",
                        keywords=["new", "overriding", "keywords"],
                    )

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.update_flow

            .. tab-item:: API Info

                .. extdoclink:: Update Flow
                    :service: flows
                    :ref: Flows/paths/~1flows~1{flow_id}/put
        """  # noqa E501

        data = {
            "title": title,
            "definition": definition,
            "input_schema": input_schema,
            "subtitle": subtitle,
            "description": description,
            "flow_owner": flow_owner,
            "flow_viewers": flow_viewers,
            "flow_starters": flow_starters,
            "flow_administrators": flow_administrators,
            "run_managers": run_managers,
            "run_monitors": run_monitors,
            "keywords": keywords,
            "subscription_id": subscription_id,
            "authentication_policy_id": authentication_policy_id,
            **(additional_fields or {}),
        }
        return self.put(f"/flows/{flow_id}", data=data)

    def delete_flow(
        self,
        flow_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """Delete a flow

        :param flow_id: The ID of the flow to delete
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Delete Flow
                    :service: flows
                    :ref: Flows/paths/~1flows~1{flow_id}/delete
        """
        return self.delete(f"/flows/{flow_id}", query_params=query_params)

    def validate_flow(
        self,
        definition: dict[str, t.Any],
        input_schema: dict[str, t.Any] | MissingType = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Validate a flow

        :param definition: JSON object specifying flow states and execution order. For
            a more detailed explanation of the flow definition, see
            `Authoring Flows <https://docs.globus.org/api/flows/authoring-flows>`_
        :param input_schema: A JSON Schema to which flow run input must conform

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    ...
                    flows = FlowsClient(...)
                    flows.validate_flow(
                        definition={
                            "StartAt": "the-one-true-state",
                            "States": {"the-one-true-state": {"Type": "Pass", "End": True}},
                        },
                        input_schema={
                            "type": "object",
                            "properties": {
                                "input-a": {"type": "string"},
                                "input-b": {"type": "number"},
                                "input-c": {"type": "boolean"},
                            },
                        },
                    )

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.validate_flow

            .. tab-item:: API Info

                .. extdoclink:: Validate Flow
                    :service: flows
                    :ref: Flows/paths/~1flows~1validate/post
        """  # noqa E501

        data = {
            "definition": definition,
            "input_schema": input_schema,
        }
        return self.post("/flows/validate", data=data)

    @paging.has_paginator(paging.MarkerPaginator, items_key="runs")
    def list_runs(
        self,
        *,
        filter_flow_id: (
            t.Iterable[uuid.UUID | str] | uuid.UUID | str | MissingType
        ) = MISSING,
        filter_roles: str | t.Iterable[str] | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableRunsResponse:
        """
        List all runs.

        :param filter_flow_id: One or more flow IDs used to filter the results
        :param filter_roles: A list of role names used to filter the results
        :param marker: A pagination marker, used to get the next page of results.
        :param query_params: Any additional parameters to be passed through

        **Filter Roles Values**

        The valid values for ``role`` are:

          - ``run_owner``
          - ``run_manager``
          - ``run_monitor``
          - ``flow_run_manager``
          - ``flow_run_monitor``

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    flows = globus_sdk.FlowsClient(...)
                    for run in flows.list_runs():
                        print(run["run_id"])

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.list_runs

            .. tab-item:: API Info

                .. extdoclink:: List Runs
                    :service: flows
                    :ref: Runs/paths/~1runs/get
        """
        query_params = {
            "filter_flow_id": commajoin(filter_flow_id),
            "filter_roles": commajoin(filter_roles),
            "marker": marker,
            **(query_params or {}),
        }
        return IterableRunsResponse(self.get("/runs", query_params=query_params))

    @paging.has_paginator(paging.MarkerPaginator, items_key="entries")
    def get_run_logs(
        self,
        run_id: uuid.UUID | str,
        *,
        limit: int | MissingType = MISSING,
        reverse_order: bool | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableRunLogsResponse:
        """
        Retrieve the execution logs associated with a run

        These logs describe state transitions and associated payloads for a run

        :param run_id: Run ID to retrieve logs for
        :param limit: Maximum number of log entries to return (server default: 10)
             (value between 1 and 100 inclusive)
        :param reverse_order: Return results in reverse chronological order (server
            default: false)
        :param marker: Marker for the next page of results (provided by the server)
        :param query_params: Any additional parameters to be passed through

        .. tab-set::

            .. tab-item:: Paginated Usage

                .. paginatedusage:: get_run_logs

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.get_run_logs

            .. tab-item:: API Info

                .. extdoclink:: Get Run Logs
                    :service: flows
                    :ref: Runs/paths/~1runs~1{action_id}~1log/get
        """

        query_params = {
            "limit": limit,
            "reverse_order": reverse_order,
            "marker": marker,
            **(query_params or {}),
        }
        return IterableRunLogsResponse(
            self.get(f"/runs/{run_id}/log", query_params=query_params)
        )

    def get_run(
        self,
        run_id: uuid.UUID | str,
        *,
        include_flow_description: bool | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        Retrieve information about a particular run of a flow

        :param run_id: The ID of the run to get
        :param include_flow_description: If set to true, the lookup will attempt to
           attach metadata about the flow to the run to the run response under the key
           "flow_description" (default: False)
        :param query_params: Any additional parameters to be passed through


        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    flows.get_run("581753c7-45da-43d3-ad73-246b46e7cb6b")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.get_run

            .. tab-item:: API Info

                .. extdoclink:: Get Run
                    :service: flows
                    :ref: Flows/paths/~1runs~1{run_id}/get
        """
        query_params = {
            "include_flow_description": include_flow_description,
            **(query_params or {}),
        }
        return self.get(f"/runs/{run_id}", query_params=query_params)

    def get_run_definition(
        self,
        run_id: uuid.UUID | str,
    ) -> GlobusHTTPResponse:
        """
        Get the flow definition and input schema at the time the run was started.

        :param run_id: The ID of the run to get

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    flows.get_run_definition("581753c7-45da-43d3-ad73-246b46e7cb6b")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.get_run_definition

            .. tab-item:: API Info

                .. extdoclink:: Get Run Definition
                    :service: flows
                    :ref: Flows/paths/~1runs~1{run_id}~1definition/get
        """

        return self.get(f"/runs/{run_id}/definition")

    def cancel_run(self, run_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Cancel a run.

        :param run_id: The ID of the run to cancel


        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    flows.cancel_run("581753c7-45da-43d3-ad73-246b46e7cb6b")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.cancel_run

            .. tab-item:: API Info

                .. extdoclink:: Cancel Run
                    :service: flows
                    :ref: Runs/paths/~1runs~1{run_id}~1cancel/post
        """

        return self.post(f"/runs/{run_id}/cancel")

    def update_run(
        self,
        run_id: uuid.UUID | str,
        *,
        label: str | MissingType = MISSING,
        tags: list[str] | MissingType = MISSING,
        run_monitors: list[str] | MissingType = MISSING,
        run_managers: list[str] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        Update the metadata of a specific run.

        :param run_id: The ID of the run to update
        :param label: A short human-readable title (1 - 64 chars)
        :param tags: A collection of searchable tags associated with the run.
            Tags are normalized by stripping leading and trailing whitespace,
            and replacing all whitespace with a single space.
        :param run_monitors: A list of authenticated entities (identified by URN)
            authorized to view this run in addition to the run owner
        :param run_managers: A list of authenticated entities (identified by URN)
            authorized to view & cancel this run in addition to the run owner
        :param additional_fields: Additional Key/Value pairs sent to the run API
            (this parameter is used to bypass local sdk key validation helping)


        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    flows.update_run(
                        "581753c7-45da-43d3-ad73-246b46e7cb6b",
                        label="Crunch numbers for experiment xDA202-batch-10",
                    )

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.update_run

            .. tab-item:: API Info

                .. extdoclink:: Update Run
                    :service: flows
                    :ref: Runs/paths/~1runs~1{run_id}/put
        """

        data = {
            "tags": tags,
            "label": label,
            "run_monitors": run_monitors,
            "run_managers": run_managers,
            **(additional_fields or {}),
        }
        return self.put(f"/runs/{run_id}", data=data)

    def delete_run(self, run_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Delete a run.

        :param run_id: The ID of the run to delete


        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    flows.delete_run("581753c7-45da-43d3-ad73-246b46e7cb6b")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.delete_run

            .. tab-item:: API Info

                .. extdoclink:: Delete Run
                    :service: flows
                    :ref: Runs/paths/~1runs~1{run_id}~1release/post
        """

        return self.post(f"/runs/{run_id}/release")

    def get_registered_api(
        self,
        registered_api_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        Retrieve a registered API by ID.

        :param registered_api_id: The ID of the registered API to fetch
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    flows.get_registered_api("a1b2c3d4-e5f6-7890-abcd-ef1234567890")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.get_registered_api

            .. tab-item:: API Info

                .. extdoclink:: Get Registered API
                    :service: flows
                    :ref: Registered APIs/paths/~1registered_apis~1
                          {registered_api_id}/get
        """
        return self.get(
            f"/registered_apis/{registered_api_id}", query_params=query_params
        )

    @paging.has_paginator(paging.MarkerPaginator, items_key="registered_apis")
    def list_registered_apis(
        self,
        *,
        filter_roles: str | t.Iterable[str] | MissingType = MISSING,
        orderby: str | t.Iterable[str] | MissingType = MISSING,
        per_page: int | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableRegisteredAPIsResponse:
        """
        List registered APIs.

        :param filter_roles: Role names to filter results (owner, administrator, viewer)
        :param orderby: Field and order for sorting results
        :param per_page: Number of results per page
        :param marker: Pagination marker for continuing results
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import FlowsClient

                    flows = FlowsClient(...)
                    for api in flows.list_registered_apis(filter_roles="owner"):
                        print(f"API: {api['name']}")

            .. tab-item:: Paginated Usage

                .. paginatedusage:: list_registered_apis

            .. tab-item:: API Info

                .. extdoclink:: List Registered APIs
                    :service: flows
                    :ref: Registered APIs/paths/~1registered_apis/get
        """
        query_params = {
            "filter_roles": commajoin(filter_roles),
            "orderby": (
                orderby if isinstance(orderby, (str, MissingType)) else list(orderby)
            ),
            "marker": marker,
            "per_page": per_page,
            **(query_params or {}),
        }
        return IterableRegisteredAPIsResponse(
            self.get("/registered_apis", query_params=query_params)
        )


class SpecificFlowClient(client.BaseClient):
    r"""
    Client for interacting with a specific flow through the Globus Flows API.

    Unlike other client types, this must be provided with a specific flow id. All other
        arguments are the same as those for :class:`~globus_sdk.BaseClient`.

    .. sdk-sphinx-copy-params:: BaseClient

        :param flow_id: The generated UUID associated with a flow

    .. automethodlist:: globus_sdk.SpecificFlowClient
    """

    error_class = FlowsAPIError
    service_name = "flows"
    scopes: SpecificFlowScopes = SpecificFlowScopes._build_class_stub()

    def __init__(
        self,
        flow_id: uuid.UUID | str,
        *,
        environment: str | None = None,
        app: GlobusApp | None = None,
        app_scopes: list[Scope] | None = None,
        authorizer: GlobusAuthorizer | None = None,
        app_name: str | None = None,
        transport: RequestsTransport | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        self._flow_id = flow_id
        self.scopes = SpecificFlowScopes(flow_id)
        super().__init__(
            app=app,
            app_scopes=app_scopes,
            environment=environment,
            authorizer=authorizer,
            app_name=app_name,
            transport=transport,
            retry_config=retry_config,
        )

    @property
    def default_scope_requirements(self) -> list[Scope]:
        return [self.scopes.user]

    def add_app_transfer_data_access_scope(
        self, collection_ids: uuid.UUID | str | t.Iterable[uuid.UUID | str]
    ) -> Self:
        """
        Add a dependent ``data_access`` scope for one or more given ``collection_ids``
        to this client's ``GlobusApp``, under the Transfer ``all`` scope.
        Useful for preventing ``ConsentRequired`` errors when starting or resuming runs
        of flows that use Globus Connect Server mapped collection(s).

        .. warning::

            This method must only be used on ``collection_ids`` for non-High-Assurance
            GCS Mapped Collections.

            Use on other collection types, e.g., on GCP Mapped Collections or any form
            of Guest Collection, will result in "Unknown Scope" errors during the login
            flow.

        Returns ``self`` for chaining.

        Raises ``GlobusSDKUsageError`` if this client was not initialized with an app.

        :param collection_ids: a collection ID or an iterable of IDs.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    flow_id = ...
                    COLLECTION_ID = ...
                    app = UserApp("myapp", client_id=NATIVE_APP_CLIENT_ID)
                    client = SpecificFlowClient(FLOW_ID, app=app).add_app_transfer_data_access_scope(
                        COLLECTION_ID
                    )

                    client.run_flow({"collection": COLLECTION_ID})
        """  # noqa: E501
        if isinstance(collection_ids, (str, uuid.UUID)):
            guards.validators.uuidlike("collection_ids", collection_ids)
            # wrap the collection_ids input in a list for consistent iteration below
            collection_ids_ = [collection_ids]
        else:
            # copy to a list so that ephemeral iterables can be iterated multiple times
            collection_ids_ = list(collection_ids)
            for i, c in enumerate(collection_ids_):
                guards.validators.uuidlike(f"collection_ids[{i}]", c)

        transfer_scope = TransferScopes.all.with_optional(True)
        for coll_id in collection_ids_:
            data_access_scope = GCSCollectionScopes(
                str(coll_id)
            ).data_access.with_optional(True)
            transfer_scope = transfer_scope.with_dependency(data_access_scope)

        specific_flow_scope = self.scopes.user.with_dependency(transfer_scope)
        self.add_app_scope(specific_flow_scope)
        return self

    def run_flow(
        self,
        body: dict[str, t.Any],
        *,
        label: str | MissingType = MISSING,
        tags: list[str] | MissingType = MISSING,
        activity_notification_policy: (
            dict[str, t.Any] | RunActivityNotificationPolicy | MissingType
        ) = MISSING,
        run_monitors: list[str] | MissingType = MISSING,
        run_managers: list[str] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        :param body: The input json object handed to the first flow state. The flows
            service will validate this object against the flow's supplied input schema.
        :param label: A short human-readable title (1 - 64 chars)
        :param tags: A collection of searchable tags associated with the run. Tags are
            normalized by stripping leading and trailing whitespace, and replacing all
            whitespace with a single space.
        :param activity_notification_policy: A policy document which declares when the
            run will send notification emails. By default, notifications are only sent
            when a run status changes to ``"INACTIVE"``.
        :param run_monitors: A list of authenticated entities (identified by URN)
            authorized to view this run in addition to the run owner
        :param run_managers: A list of authenticated entities (identified by URN)
            authorized to view & cancel this run in addition to the run owner
        :param additional_fields: Additional Key/Value pairs sent to the run API
            (this parameter is used to bypass local sdk key validation helping)

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Run Flow
                    :service: flows
                    :ref: ~1flows~1{flow_id}~1run/post
        """
        data = {
            "body": body,
            "tags": tags,
            "label": label,
            "activity_notification_policy": activity_notification_policy,
            "run_monitors": run_monitors,
            "run_managers": run_managers,
            **(additional_fields or {}),
        }
        return self.post(f"/flows/{self._flow_id}/run", data=data)

    def resume_run(self, run_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        :param run_id: The ID of the run to resume

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import SpecificFlowClient

                    ...
                    flow = SpecificFlowClient(flow_id, ...)
                    flow.resume_run(run_id)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.resume_run

            .. tab-item:: API Info

                .. extdoclink:: Resume Run
                    :service: flows
                    :ref: Runs/paths/~1flows~1{flow_id}~1runs~1{run_id}~1resume/post
        """
        return self.post(f"/runs/{run_id}/resume")

    def validate_run(
        self,
        body: dict[str, t.Any],
        *,
        label: str | MissingType = MISSING,
        tags: list[str] | MissingType = MISSING,
        run_monitors: list[str] | MissingType = MISSING,
        run_managers: list[str] | MissingType = MISSING,
        activity_notification_policy: (
            dict[str, t.Any] | RunActivityNotificationPolicy | MissingType
        ) = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        :param body: The parameters to validate against the flow's input schema.
        :param label: A short human-readable title.
        :param tags: A collection of searchable tags associated with the run.
            Tags are normalized by stripping leading and trailing whitespace,
            and replacing all whitespace with a single space.
        :param run_monitors: A list of Globus Auth principals (identified by URN)
            authorized to monitor this run (in addition to the run owner).
        :param run_managers: A list of Globus Auth principals (identified by URN)
            authorized to manage this run (in addition to the run owner).
        :param activity_notification_policy:
            A policy document which declares when the Flows service will send
            notification emails regarding the run's activity.
        :param additional_fields: Additional key/value pairs sent to the run API.
            This parameter can be used to bypass SDK parameter validation.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    from globus_sdk import SpecificFlowClient

                    ...
                    flow = SpecificFlowClient(flow_id, ...)
                    flow.validate_run(body={"param": "value"})
            .. tab-item:: Example Response Data

                .. expandtestfixture:: flows.validate_run
        """

        data = {
            "body": body,
            "tags": tags,
            "label": label,
            "run_monitors": run_monitors,
            "run_managers": run_managers,
            "activity_notification_policy": activity_notification_policy,
        }
        data.update(additional_fields or {})

        return self.post(f"/flows/{self._flow_id}/validate_run", data=data)



from globus_sdk import LocalGlobusConnectServer


def test_info_dict_from_nonexistent_file_is_none(tmp_path):
    info_path = tmp_path / "info.json"
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is None


def test_info_dict_from_non_json_file_is_none(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_text("{")
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is None


def test_info_dict_from_non_json_object_file_is_none(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_text("[]")
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is None


def test_info_dict_from_non_unicode_file_is_none(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_bytes(b'{"foo":"{' + bytes.fromhex("1BAD DEC0DE") + b'}"}')
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is None


def test_info_dict_from_empty_json_file_is_okay_but_has_no_properties(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_text("{}")
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is not None
    assert gcs.endpoint_id is None
    assert gcs.domain_name is None


def test_info_dict_can_load_endpoint_id(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_text('{"endpoint_id": "foo"}')
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is not None
    assert gcs.endpoint_id == "foo"


def test_info_dict_can_load_domain(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_text('{"domain_name": "foo"}')
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is not None
    assert gcs.domain_name == "foo"


def test_endpoint_id_property_ignores_non_str_value(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_text('{"endpoint_id": 1}')
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is not None
    assert gcs.endpoint_id is None


def test_domain_name_property_ignores_non_str_value(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_text('{"domain_name": ["foo"]}')
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict is not None
    assert gcs.domain_name is None


def test_info_dict_can_reload_with_deleter(tmp_path):
    info_path = tmp_path / "info.json"
    info_path.write_text('{"foo": "bar"}')

    # initial load okay
    gcs = LocalGlobusConnectServer(info_path=info_path)
    assert gcs.info_dict == {"foo": "bar"}
    # update the data on disk
    info_path.write_text('{"bar": "baz"}')
    # data unchanged on the instance
    assert gcs.info_dict == {"foo": "bar"}
    # clear and reload updates the data
    del gcs.info_dict
    assert gcs.info_dict == {"bar": "baz"}



from __future__ import annotations

import dataclasses
import re
import typing as t
import uuid

_NORMALIZATION_PATTERN = re.compile(r"[_\- ]+")


def _normalize_name(name: str) -> str:
    return _NORMALIZATION_PATTERN.sub("-", name.strip()).lower()


@dataclasses.dataclass
class GlobusConnectServerConnector:
    """
    A container for Globus Connect Server Connector descriptions.
    Contains a ``name`` and a ``connector_id``.
    """

    name: str
    connector_id: str


class ConnectorTable:
    """
    This class defines the known Globus Connect Server Connectors in a mapping
    structure.

    It supports access by attribute or via a helper method for doing lookups.
    For example, all of the following three usages retrieve the Azure Blob connector:

    .. code-block:: pycon

        >>> ConnectorTable.AZURE_BLOB
        >>> ConnectorTable.lookup("Azure Blob")
        >>> ConnectorTable.lookup("9436da0c-a444-11eb-af93-12704e0d6a4d")

    Given the results of such a lookup, you can retrieve the canonical name and ID for
    a connector like so:

    .. code-block:: pycon

        >>> connector = ConnectorTable.AZURE_BLOB
        >>> connector.name
        'Azure Blob'
        >>> connector.connector_id
        '9436da0c-a444-11eb-af93-12704e0d6a4d'
    """

    _connectors: t.ClassVar[tuple[tuple[str, str, str], ...]] = (
        ("ACTIVESCALE", "ActiveScale", "7251f6c8-93c9-11eb-95ba-12704e0d6a4d"),
        ("AZURE_BLOB", "Azure Blob", "9436da0c-a444-11eb-af93-12704e0d6a4d"),
        ("BLACKPEARL", "BlackPearl", "7e3f3f5e-350c-4717-891a-2f451c24b0d4"),
        ("BOX", "Box", "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"),
        ("CEPH", "Ceph", "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72"),
        ("DROPBOX", "Dropbox", "49b00fd6-63f1-48ae-b27f-d8af4589f876"),
        (
            "GOOGLE_CLOUD_STORAGE",
            "Google Cloud Storage",
            "56366b96-ac98-11e9-abac-9cb6d0d9fd63",
        ),
        ("GOOGLE_DRIVE", "Google Drive", "976cf0cf-78c3-4aab-82d2-7c16adbcc281"),
        ("HPSS", "HPSS", "fb656a17-0f69-4e59-95ff-d0a62ca7bdf5"),
        ("IRODS", "iRODS", "e47b6920-ff57-11ea-8aaa-000c297ab3c2"),
        ("POSIX", "POSIX", "145812c8-decc-41f1-83cf-bb2a85a2a70b"),
        ("POSIX_STAGING", "POSIX Staging", "052be037-7dda-4d20-b163-3077314dc3e6"),
        ("ONEDRIVE", "OneDrive", "28ef55da-1f97-11eb-bdfd-12704e0d6a4d"),
        ("S3", "S3", "7643e831-5f6c-4b47-a07f-8ee90f401d23"),
    )

    ACTIVESCALE: t.ClassVar[GlobusConnectServerConnector]
    AZURE_BLOB: t.ClassVar[GlobusConnectServerConnector]
    BLACKPEARL: t.ClassVar[GlobusConnectServerConnector]
    BOX: t.ClassVar[GlobusConnectServerConnector]
    CEPH: t.ClassVar[GlobusConnectServerConnector]
    DROPBOX: t.ClassVar[GlobusConnectServerConnector]
    GOOGLE_CLOUD_STORAGE: t.ClassVar[GlobusConnectServerConnector]
    GOOGLE_DRIVE: t.ClassVar[GlobusConnectServerConnector]
    HPSS: t.ClassVar[GlobusConnectServerConnector]
    IRODS: t.ClassVar[GlobusConnectServerConnector]
    ONEDRIVE: t.ClassVar[GlobusConnectServerConnector]
    POSIX: t.ClassVar[GlobusConnectServerConnector]
    POSIX_STAGING: t.ClassVar[GlobusConnectServerConnector]
    S3: t.ClassVar[GlobusConnectServerConnector]

    @classmethod
    def all_connectors(cls) -> t.Iterable[GlobusConnectServerConnector]:
        """
        Return an iterator of all known connectors.
        """
        for attribute, _, _ in cls._connectors:
            item: GlobusConnectServerConnector = getattr(cls, attribute)
            yield item

    @classmethod
    def lookup(cls, name_or_id: uuid.UUID | str) -> GlobusConnectServerConnector | None:
        """
        Convert a name or ID into a connector object.
        Returns None if the name or ID is not recognized.

        Names are normalized before lookup so that they are case-insensitive and
        spaces, dashes, and underscores are all treated equivalently. For
        example, ``Google Drive``, ``google-drive``, and ``gOOgle_dRiVe`` are
        all equivalent.

        :param name_or_id: The name or ID of the connector
        """
        normalized = _normalize_name(str(name_or_id))
        for connector in cls.all_connectors():
            if normalized == connector.connector_id or normalized == _normalize_name(
                connector.name
            ):
                return connector
        return None

    @classmethod
    def extend(
        cls,
        *,
        connector_name: str,
        connector_id: uuid.UUID | str,
        attribute_name: str | None = None,
    ) -> type[ConnectorTable]:
        """
        Extend the ConnectorTable class with a new connector, returning a new
        ConnectorTable subclass.

        Usage example:

        .. code-block:: pycon

            >>> MyTable = ConnectorTable.extend(
            ...     connector_name="Star Trek Transporter",
            ...     connector_id="0b19772d-729a-4c8f-93e1-59d5778cecf3",
            ... )
            >>> obj = MyTable.STAR_TREK_TRANSPORTER
            >>> obj.connector_id
            '0b19772d-729a-4c8f-93e1-59d5778cecf3'
            >>> obj.name
            'Star Trek Transporter'

        :param connector_name: The name of the connector to add
        :param connector_id: The ID of the connector to add
        :param attribute_name: The attribute name with which the connector will be
            attached to the new subclass. Defaults to the connector name uppercased and
            with spaces converted to underscores.
        """
        if attribute_name is None:
            attribute_name = connector_name.upper().replace(" ", "_")
        connector_id_str = str(connector_id)

        connectors = cls._connectors + (
            (attribute_name, connector_name, connector_id_str),
        )
        connector_obj = GlobusConnectServerConnector(
            name=connector_name, connector_id=connector_id_str
        )
        return type(
            "ExtendedConnectorTable",
            (cls,),
            {"_connectors": connectors, attribute_name: connector_obj},
        )


# "render" the _connectors to live attributes of the ConnectorTable
for _attribute, _name, _id in ConnectorTable._connectors:
    setattr(
        ConnectorTable,
        _attribute,
        GlobusConnectServerConnector(name=_name, connector_id=_id),
    )
del _attribute, _name, _id



import globus_sdk

ac = globus_sdk.AuthClient()

# create new policy with keyword-only args, as supported
ac.create_policy(
    project_id="foo",
    display_name="My Policy",
    description="This is a policy",
)

# create using positional args (deprecated/unsupported)
ac.create_policy(  # type: ignore[misc]
    "foo",
    True,  # type: ignore[arg-type]
    101,  # type: ignore[arg-type]
    "My Policy",  # type: ignore[arg-type]
    "This is a policy",  # type: ignore[arg-type]
)
